#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import copy
import datetime
import io
import json
import logging
import os
import time
import urllib.parse
from typing import Any, Dict, List, Optional

import fsspec
import oci.data_flow
import oci.util as oci_util
import yaml
from ads.common import utils
from ads.common.auth import default_signer
from ads.common.decorator.utils import class_or_instance_method
from ads.common.oci_client import OCIClientFactory
from ads.common.oci_mixin import OCIModelMixin
from ads.common.utils import batch_convert_case, camel_to_snake
from ads.jobs.builders.infrastructure.base import Infrastructure, RunInstance
from ads.jobs.builders.infrastructure.utils import normalize_config
from ads.jobs.builders.runtimes.python_runtime import DataFlowRuntime
from ads.model.runtime.env_info import InferenceEnvInfo
from oci.data_flow.models import CreateApplicationDetails, CreateRunDetails
from tqdm import tqdm

logger = logging.getLogger(__name__)

CONDA_PACK_SUFFIX = "#conda"
SLEEP_INTERVAL = 3


def conda_pack_name_to_dataflow_config(conda_uri):
    return {
        "spark.archives": conda_uri + CONDA_PACK_SUFFIX,
        "dataflow.auth": "resource_principal",
    }


def _env_variables_to_dataflow_config(
    env_vars: Dict[str, str] = None
) -> Dict[str, str]:
    """Prepares environment variables for the application.
    Similar environment variables will be setup for the driver and executor.

    Parameters
    ----------
    env_vars: (Dict[str, str], optional). Defaults to `None`
        The dictionary with SRC env variables.
        Example: {"env1": "value1"}

    Returns
    -------
    Dict[str, str]
        Dictionary with pre populated environment variables.
        Example: {"spark.executorEnv.env1": "value1", "spark.driverEnv.env1": "value1"}
    """
    if not env_vars:
        return {}

    result = {}
    for level in ("spark.executorEnv", "spark.driverEnv"):
        for env_name, env_value in env_vars.items():
            result[f"{level}.{env_name}"] = env_value

    return result


class DataFlowApp(OCIModelMixin, oci.data_flow.models.Application):
    @classmethod
    def init_client(cls, **kwargs) -> oci.data_flow.data_flow_client.DataFlowClient:
        return cls._init_client(
            client=oci.data_flow.data_flow_client.DataFlowClient, **kwargs
        )

    @property
    def client(self) -> oci.data_flow.data_flow_client.DataFlowClient:
        return super().client

    def create(self) -> "DataFlowApp":
        """
        Create a Data Flow application.

        Returns
        -------
        DataFlowApp
            a DataFlowApp instance

        """
        resp = self.client.create_application(
            self.to_oci_model(CreateApplicationDetails)
        )
        logger.debug(f"Created a DataFlow Application {resp.data}")
        for k, v in json.loads(repr(resp.data)).items():
            setattr(self, k, v)
        return self

    def delete(self) -> None:
        """
        Delete a Data Flow application.

        Returns
        -------
        None
        """
        self.application = self.client.delete_application(self.id)
        self.lifecycle_state = oci.data_flow.models.Application.LIFECYCLE_STATE_DELETED

    def __repr__(self) -> str:
        """Displays the object as YAML."""
        return self.to_yaml()

    def to_yaml(self) -> str:
        """Serializes the object into YAML string.

        Returns
        -------
        str
            YAML stored in a string.
        """
        return yaml.safe_dump(self.to_dict())


class DataFlowRun(OCIModelMixin, oci.data_flow.models.Run, RunInstance):
    _DETAILS_LINK = (
        "https://console.{region}.oraclecloud.com/data-flow/runs/details/{id}"
    )

    TERMINATED_STATES = [
        oci.data_flow.models.Run.LIFECYCLE_STATE_CANCELED,
        oci.data_flow.models.Run.LIFECYCLE_STATE_FAILED,
        oci.data_flow.models.Run.LIFECYCLE_STATE_SUCCEEDED,
    ]

    @classmethod
    def init_client(cls, **kwargs) -> oci.data_flow.data_flow_client.DataFlowClient:
        return cls._init_client(
            client=oci.data_flow.data_flow_client.DataFlowClient, **kwargs
        )

    @property
    def client(self) -> oci.data_flow.data_flow_client.DataFlowClient:
        return super().client

    def create(self) -> "DataFlowRun":
        """
        Create a Data Flow run.

        Returns
        -------
        DataFlowRun
            a DataFlowRun instance
        """
        resp = self.client.create_run(self.to_oci_model(CreateRunDetails))
        logger.debug(f"Created a DataFlow Run {resp.data}")
        for k, v in json.loads(repr(resp.data)).items():
            setattr(self, k, v)
        return self

    @property
    def status(self) -> str:
        """
        Show status (lifecycle state) of a run.

        Returns
        -------
        str
            status of the run
        """
        if self.id:
            resp = self.client.get_run(self.id)
            self.lifecycle_state = resp.data.lifecycle_state
        return self.lifecycle_state

    @property
    def logs(self) -> "DataFlowLogs":
        """
        Show logs from a run.
        There are three types of logs: application log, driver log and executor log,
        each with stdout and stderr separately.
        To access each type of logs,
        >>> dfr.logs.application.stdout
        >>> dfr.logs.driver.stderr

        Returns
        -------
        DataFlowLogs
            an instance of DataFlowLogs
        """
        return DataFlowLogs(run_id=self.id)

    def wait(self, interval: int = SLEEP_INTERVAL) -> "DataFlowRun":
        """
        Wait for a run to terminate.

        Parameters
        ----------
        interval: int, optional
            interval to wait before probing again

        Returns
        -------
        DataFlowRun
            a DataFlowRun instance
        """
        current = self.status
        timestamp = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
        logger.info(f"{timestamp} - {self.status}")
        print(f"{timestamp} - {self.status}")
        while current not in self.TERMINATED_STATES:
            time.sleep(interval)
            if self.status != current:
                timestamp = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
                logger.info(f"{timestamp} - {self.status}")
                print(f"{timestamp} - {self.status}")
                current = self.status
        return self

    def watch(self, interval: int = SLEEP_INTERVAL) -> "DataFlowRun":
        """This is an alias of `wait()` method. It waits for a run to terminate.

        Parameters
        ----------
        interval: int, optional
            interval to wait before probing again

        Returns
        -------
        DataFlowRun
            a DataFlowRun instance
        """
        logger.info(
            "This is an alias of `wait`. Logs are available after run completes."
        )
        return self.wait(interval=interval)

    def to_yaml(self) -> str:
        """Serializes the object into YAML string.

        Returns
        -------
        str
            YAML stored in a string.
        """
        run = self.to_dict()
        run["lifecycleState"] = self.status
        return yaml.safe_dump(run)

    def delete(self) -> "DataFlowRun":
        """
        Cancel and delete a Data Flow run if it is not yet terminated.
        Will be executed asynchronously.

        Returns
        -------
        self
            The dataflow run instance.
        """
        if self.status not in self.TERMINATED_STATES:
            self.client.delete_run(self.id)
            self.lifecycle_state = oci.data_flow.models.Run.LIFECYCLE_STATE_CANCELING

        return self

    def cancel(self) -> "DataFlowRun":
        """Cancel a Data Flow run if it is not yet terminated.
        Will be executed synchronously.

        Returns
        -------
        self
            The dataflow run instance.
        """
        self.delete()
        self.wait()
        return self


class _Log:
    def __init__(self, run_id, src_type, output_type):
        self.run_id = run_id
        self.src_type = src_type
        self.output_type = output_type
        self._cache = None

    @property
    def client(self):
        return OCIClientFactory(**default_signer()).dataflow

    def _get_logs(self, n=None, reverse=False):
        if not self._cache:
            logs = self.client.list_run_logs(self.run_id).data
            lines = []
            for log in logs:
                if log.type == self.output_type and self.src_type == log.source:
                    lines.extend(
                        self.client.get_run_log(self.run_id, log.name)
                        .data.text.lstrip("\x00")
                        .rstrip("\n")
                        .splitlines()
                    )
            self._cache = lines
        if n and reverse:
            return "\n".join(self._cache[-n:])
        elif n and not reverse:
            return "\n".join(self._cache[:n])
        else:
            return "\n".join(self._cache)

    def __repr__(self):
        return self._get_logs()

    def head(self, n=None):
        return self._get_logs(n=n)

    def tail(self, n=None):
        return self._get_logs(n=n, reverse=True)

    def download(self, path):
        with open(path, mode="w") as f:
            f.write(self._get_logs())


class _DataFlowLog:
    def __init__(self, run_id, src_type):
        self.run_id = run_id
        self.src_type = src_type

    @property
    def stdout(self):
        return _Log(self.run_id, self.src_type, "STDOUT")

    @property
    def stderr(self):
        return _Log(self.run_id, self.src_type, "STDERR")


class DataFlowLogs:
    def __init__(self, run_id):
        self.run_id = run_id

    @property
    def application(self):
        return _DataFlowLog(run_id=self.run_id, src_type="APPLICATION")

    @property
    def driver(self):
        return _DataFlowLog(run_id=self.run_id, src_type="DRIVER")

    @property
    def executor(self):
        return _DataFlowLog(run_id=self.run_id, src_type="EXECUTOR")


class DataFlow(Infrastructure):

    CONST_COMPARTMENT_ID = "compartment_id"
    CONST_CONFIG = "configuration"
    CONST_EXECUTE = "execute"
    CONST_DRIVER_SHAPE = "driver_shape"
    CONST_EXECUTOR_SHAPE = "executor_shape"
    CONST_LANGUAGE = "language"
    CONST_METASTORE_ID = "metastore_id"
    CONST_BUCKET_URI = "logs_bucket_uri"
    CONST_NUM_EXECUTORS = "num_executors"
    CONST_SPARK_VERSION = "spark_version"
    CONST_WAREHOUSE_BUCKET_URI = "warehouse_bucket_uri"
    CONST_DRIVER_SHAPE_CONFIG = "driver_shape_config"
    CONST_EXECUTOR_SHAPE_CONFIG = "executor_shape_config"
    CONST_MEMORY_IN_GBS = "memory_in_gbs"
    CONST_OCPUS = "ocpus"
    CONST_ID = "id"
    CONST_PRIVATE_ENDPOINT_ID = "private_endpoint_id"

    attribute_map = {
        CONST_COMPARTMENT_ID: "compartmentId",
        CONST_CONFIG: CONST_CONFIG,
        CONST_EXECUTE: CONST_EXECUTE,
        CONST_DRIVER_SHAPE: "driverShape",
        CONST_EXECUTOR_SHAPE: "executorShape",
        CONST_METASTORE_ID: "metastoreId",
        CONST_BUCKET_URI: "logsBucketUri",
        CONST_NUM_EXECUTORS: "numExecutors",
        CONST_SPARK_VERSION: "sparkVersion",
        CONST_WAREHOUSE_BUCKET_URI: "warehouseBucketUri",
        CONST_DRIVER_SHAPE_CONFIG: "driverShapeConfig",
        CONST_EXECUTOR_SHAPE_CONFIG: "executorShapeConfig",
        CONST_MEMORY_IN_GBS: "memoryInGBs",
        CONST_OCPUS: CONST_OCPUS,
        CONST_ID: CONST_ID,
        CONST_PRIVATE_ENDPOINT_ID: "privateEndpointId",
    }

    def __init__(self, spec: dict = None, **kwargs):
        defaults = self._load_default_properties()
        spec = self._standardize_spec(spec)
        kwargs = self._standardize_spec(kwargs)
        if spec is None:
            super(DataFlow, self).__init__(defaults, **kwargs)
        else:
            spec = {
                k: v
                for k, v in spec.items()
                if f"with_{camel_to_snake(k)}" in self.__dir__() and v is not None
            }
            defaults.update(spec)
            super().__init__(defaults, **kwargs)

        self.df_app = DataFlowApp(**self._spec)
        self.runtime = None
        self._name = None

    @staticmethod
    def _load_default_properties() -> dict:
        """
        Load default properties from environment variables, notebook session, etc.

        Returns
        -------
        dict
            a dictionary of default properties
        """
        defaults = {}
        if "NB_SESSION_COMPARTMENT_OCID" in os.environ:
            defaults["compartment_id"] = os.environ["NB_SESSION_COMPARTMENT_OCID"]
        if "NB_SESSION_OCID" in os.environ:
            dsc_client = OCIClientFactory(**default_signer()).data_science
            try:
                nb_session = dsc_client.get_notebook_session(
                    os.environ["NB_SESSION_OCID"]
                ).data
                nb_config = nb_session.notebook_session_configuration_details
                defaults["driver_shape"] = nb_config.shape
                logger.debug(f"Set driver shape to {nb_config.shape}")
                defaults["executor_shape"] = nb_config.shape
                logger.debug(f"Set executor shape to {nb_config.shape}")
                if nb_config.notebook_session_shape_config_details:
                    notebook_shape_config_details = oci_util.to_dict(
                        nb_config.notebook_session_shape_config_details
                    )
                    defaults["driver_shape_config"] = copy.deepcopy(
                        notebook_shape_config_details
                    )
                    logger.debug(
                        f"Set driver shape config to {nb_config.notebook_session_shape_config_details}"
                    )
                    defaults["executor_shape_config"] = copy.deepcopy(
                        notebook_shape_config_details
                    )
                    logger.debug(
                        f"Set executor shape config to {nb_config.notebook_session_shape_config_details}"
                    )

            except Exception as e:
                logger.warning(
                    f"Error fetching details about Notebook session: {os.environ['NB_SESSION_OCID']}. {e}"
                )

        defaults["language"] = "PYTHON"
        defaults["spark_version"] = "3.2.1"
        defaults["num_executors"] = 1
        logger.debug("Set spark version to be 3.2.1.")
        logger.debug("Set number of executors to be 1.")
        return defaults

    @property
    def name(self) -> str:
        """Display name of the job"""
        return self._name

    @name.setter
    def name(self, value: str):
        """Sets the display name of the job

        Parameters
        ----------
        value : str
            The display name of the job
        """
        if self.df_app:
            self.df_app.display_name = value
        self._name = value

    @property
    def job_id(self) -> Optional[str]:
        """The OCID of the job"""
        return self.get_spec("id")

    def with_compartment_id(self, id: str) -> "DataFlow":
        """
        Set compartment id for a Data Flow job.

        Parameters
        ----------
        id: str
            compartment id

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """
        return self.set_spec(self.CONST_COMPARTMENT_ID, id)

    def with_configuration(self, configs: dict) -> "DataFlow":
        """
        Set configuration for a Data Flow job.

        Parameters
        ----------
        configs: dict
            dictionary of configurations

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """
        return self.set_spec(self.CONST_CONFIG, configs)

    def with_execute(self, exec: str) -> "DataFlow":
        """
        Set command for spark-submit.

        Parameters
        ----------
        exec: str
            str of commands

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """
        return self.set_spec(self.CONST_EXECUTE, exec)

    def with_driver_shape(self, shape: str) -> "DataFlow":
        """
        Set driver shape for a Data Flow job.

        Parameters
        ----------
        shape: str
            driver shape

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """
        return self.set_spec(self.CONST_DRIVER_SHAPE, shape)

    def with_executor_shape(self, shape: str) -> "DataFlow":
        """
        Set executor shape for a Data Flow job.

        Parameters
        ----------
        shape: str
            executor shape

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """
        return self.set_spec(self.CONST_EXECUTOR_SHAPE, shape)

    def with_language(self, lang: str) -> "DataFlow":
        """
        Set language for a Data Flow job.

        Parameters
        ----------
        lang: str
            language for the job

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """
        return self.set_spec(self.CONST_LANGUAGE, lang)

    def with_metastore_id(self, id: str) -> "DataFlow":
        """
        Set Hive metastore id for a Data Flow job.

        Parameters
        ----------
        id: str
            metastore id

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """

        return self.set_spec(self.CONST_METASTORE_ID, id)

    def with_logs_bucket_uri(self, uri: str) -> "DataFlow":
        """
        Set logs bucket uri for a Data Flow job.

        Parameters
        ----------
        uri: str
            uri to logs bucket

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """
        return self.set_spec(self.CONST_BUCKET_URI, uri)

    def with_num_executors(self, n: int) -> "DataFlow":
        """
        Set number of executors for a Data Flow job.

        Parameters
        ----------
        n: int
            number of executors

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """
        return self.set_spec(self.CONST_NUM_EXECUTORS, n)

    def with_spark_version(self, ver: str) -> "DataFlow":
        """
        Set spark version for a Data Flow job.
        Currently supported versions are 2.4.4, 3.0.2 and 3.2.1
        Documentation: https://docs.oracle.com/en-us/iaas/data-flow/using/dfs_getting_started.htm#before_you_begin

        Parameters
        ----------
        ver: str
            spark version

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """
        return self.set_spec(self.CONST_SPARK_VERSION, ver)

    def with_warehouse_bucket_uri(self, uri: str) -> "DataFlow":
        """
        Set warehouse bucket uri for a Data Flow job.

        Parameters
        ----------
        uri: str
            uri to warehouse bucket

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """
        return self.set_spec(self.CONST_WAREHOUSE_BUCKET_URI, uri)

    def with_id(self, id: str) -> "DataFlow":
        """
        Set id for a Data Flow job.

        Parameters
        ----------
        id: str
            id of a job

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """
        return self.set_spec(self.CONST_ID, id)

    def with_driver_shape_config(
        self, memory_in_gbs: float, ocpus: float, **kwargs: Dict[str, Any]
    ) -> "DataFlow":
        """
        Sets the driver shape config details of Data Flow job infrastructure.
        Specify only when a flex shape is selected.
        For example `VM.Standard.E3.Flex` allows the memory_in_gbs and cpu count to be specified.

        Parameters
        ----------
        memory_in_gbs: float
            The size of the memory in GBs.
        ocpus: float
            The OCPUs count.
        kwargs
            Additional keyword arguments.

        Returns
        -------
        DataFlow
            the Data Flow instance itself.
        """
        return self.set_spec(
            self.CONST_DRIVER_SHAPE_CONFIG,
            {
                self.CONST_OCPUS: ocpus,
                self.CONST_MEMORY_IN_GBS: memory_in_gbs,
                **kwargs,
            },
        )

    def with_executor_shape_config(
        self, memory_in_gbs: float, ocpus: float, **kwargs: Dict[str, Any]
    ) -> "DataFlow":
        """
        Sets the executor shape config details of Data Flow job infrastructure.
        Specify only when a flex shape is selected.
        For example `VM.Standard.E3.Flex` allows the memory_in_gbs and cpu count to be specified.

        Parameters
        ----------
        memory_in_gbs: float
            The size of the memory in GBs.
        ocpus: float
            The OCPUs count.
        kwargs
            Additional keyword arguments.

        Returns
        -------
        DataFlow
            the Data Flow instance itself.
        """
        return self.set_spec(
            self.CONST_EXECUTOR_SHAPE_CONFIG,
            {
                self.CONST_OCPUS: ocpus,
                self.CONST_MEMORY_IN_GBS: memory_in_gbs,
                **kwargs,
            },
        )

    def with_private_endpoint_id(self, private_endpoint_id: str) -> "DataFlow":
        """
        Set the private endpoint ID for a Data Flow job infrastructure.

        Parameters
        ----------
        private_endpoint_id: str
            The OCID of a private endpoint.

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """
        return self.set_spec(self.CONST_PRIVATE_ENDPOINT_ID, private_endpoint_id)

    def __getattr__(self, item):
        if f"with_{item}" in self.__dir__():
            return self.get_spec(item)
        raise AttributeError(f"Attribute {item} not found.")

    def create(self, runtime: DataFlowRuntime, **kwargs) -> "DataFlow":
        """
        Create a Data Flow job given a runtime.

        Parameters
        ----------
        runtime
            runtime to bind to the Data Flow job
        kwargs
            additional keyword arguments

        Returns
        -------
        DataFlow
            a Data Flow job instance
        """
        # Set default display_name if not specified - randomly generated easy to remember name
        if not self.name:
            self.name = utils.get_random_name_for_resource()
        payload = copy.deepcopy(self._spec)
        runtime.convert(overwrite=kwargs.get("overwrite", False))
        if not runtime.script_uri:
            raise ValueError("script uri must be specified in runtime.")
        overwrite = kwargs.get("overwrite", False)
        if runtime.script_uri.split(":")[0] != "oci":
            if runtime.script_bucket:
                runtime.with_script_uri(
                    self._upload_file(
                        runtime.script_uri, runtime.script_bucket, overwrite
                    )
                )
            else:
                raise ValueError(
                    "script bucket must be specified if script uri given is local."
                )
        if runtime.archive_uri and runtime.archive_uri.split(":")[0] != "oci":
            if runtime.archive_bucket:
                runtime.with_archive_uri(
                    self._upload_file(
                        runtime.archive_uri, runtime.archive_bucket, overwrite
                    )
                )
            else:
                raise ValueError(
                    "archive bucket must be specified if archive uri given is local."
                )
        if runtime.conda:
            conda_type = runtime.conda.get(runtime.CONST_CONDA_TYPE)
            if conda_type == runtime.CONST_CONDA_TYPE_CUSTOM:
                conda_uri = runtime.conda.get(runtime.CONST_CONDA_URI)
            elif conda_type == runtime.CONST_CONDA_TYPE_SERVICE:
                conda_slug = runtime.conda.get(runtime.CONST_CONDA_SLUG)
                env_info = InferenceEnvInfo.from_slug(env_slug=conda_slug)
                conda_uri = env_info.inference_env_path
                # TODO: Re-visit if we can enable a global read policy
                raise NotImplementedError(
                    "Service Conda Packs not yet supported. Please use a published pack."
                )
            else:
                raise ValueError(f"Conda built type not understood: {conda_type}.")
            runtime_config = runtime.configuration or dict()
            runtime_config.update(conda_pack_name_to_dataflow_config(conda_uri))
            runtime.with_configuration(runtime_config)

        payload.update(
            {
                "display_name": self.name,
                "file_uri": runtime.script_uri,
                "freeform_tags": runtime.freeform_tags,
                "archive_uri": runtime.archive_uri,
                "configuration": runtime.configuration,
            }
        )
        if len(runtime.args) > 0:
            payload["arguments"] = runtime.args
        if not payload.get("compartment_id", None):
            raise ValueError(
                "Compartment id is required. Specify compartment id via 'with_compartment_id()'."
            )
        payload.pop("id", None)
        logger.debug(f"Creating a DataFlow Application with payload {payload}")
        self.df_app = DataFlowApp(**payload).create()
        self.with_id(self.df_app.id)
        return self

    @staticmethod
    def _upload_file(local_path, bucket, overwrite=False):
        signer = default_signer()
        os_client = OCIClientFactory(**signer).object_storage
        namespace = os_client.get_namespace().data
        parsed = urllib.parse.urlparse(bucket)
        chunk_size = io.DEFAULT_BUFFER_SIZE
        if parsed.scheme == "oci":
            dst_path = os.path.join(bucket, os.path.basename(local_path))
        else:
            dst_path = f"oci://{bucket}@{namespace}/{os.path.basename(local_path)}"
        logger.debug(f"Uploading {local_path} to {dst_path}")
        with fsspec.open(dst_path, mode="wb", **signer) as fto:
            file_system_clz = fsspec.get_filesystem_class(
                urllib.parse.urlparse(local_path).scheme or "file"
            )
            file_size = file_system_clz().info(local_path)["size"]

            if not overwrite:
                remote_file_system_clz = fsspec.get_filesystem_class(
                    urllib.parse.urlparse(dst_path).scheme or "file"
                )
                remote_file_system = remote_file_system_clz(**default_signer())
                if remote_file_system.exists(dst_path):
                    raise FileExistsError(
                        f"{dst_path} exists. Please use a new file name."
                    )

            with fsspec.open(local_path, mode="rb", encoding=None) as fread:
                with tqdm.wrapattr(
                    fread,
                    "read",
                    desc=f"Uploading file: {local_path} to {dst_path}",
                    total=file_size,
                ) as ffrom:
                    while True:
                        chunk = ffrom.read(chunk_size)
                        if chunk:
                            fto.write(chunk)
                        else:
                            break
        return dst_path

    def run(
        self,
        name: str = None,
        args: List[str] = None,
        env_vars: Dict[str, str] = None,
        freeform_tags: Dict[str, str] = None,
        wait: bool = False,
        **kwargs,
    ) -> DataFlowRun:
        """
        Run a Data Flow job.

        Parameters
        ----------
        name: str, optional
            name of the run. If a name is not provided, a randomly generated easy to remember name
            with timestamp will be generated, like 'strange-spider-2022-08-17-23:55.02'.
        args: List[str], optional
            list of command line arguments
        env_vars: Dict[str, str], optional
            dictionary of environment variables (not used for data flow)
        freeform_tags: Dict[str, str], optional
            freeform tags
        wait: bool, optional
            whether to wait for a run to terminate
        kwargs
            additional keyword arguments

        Returns
        -------
        DataFlowRun
            a DataFlowRun instance
        """
        payload = kwargs
        if "application_id" not in payload:
            payload["application_id"] = self.id
        if "compartment_id" not in payload:
            payload["compartment_id"] = self.get_spec("compartment_id")
        # Set default display_name if not specified - randomly generated easy to remember name generated
        payload["display_name"] = name if name else utils.get_random_name_for_resource()
        payload["arguments"] = args if args and len(args) > 0 else None
        payload["freeform_tags"] = freeform_tags
        payload.pop("spark_version", None)
        logger.debug(f"Creating a DataFlow Run with payload {payload}")
        run = DataFlowRun(**payload).create()
        if wait:
            run.wait(kwargs.pop("interval", SLEEP_INTERVAL))
        return run

    def run_list(self, **kwargs) -> List[DataFlowRun]:
        """
        List runs associated with a Data Flow job.

        Parameters
        ----------
        kwargs
         additional arguments for filtering runs.

        Returns
        -------
        List[DataFlowRun]
            list of DataFlowRun instances
        """
        items = oci.pagination.list_call_get_all_results(
            self.df_app.client.list_runs,
            self.compartment_id,
            application_id=self.id,
            **kwargs,
        ).data
        return [DataFlowRun.from_oci_model(item) for item in items]

    def delete(self):
        """
        Delete a Data Flow job and canceling associated runs.

        Returns
        -------
        None
        """
        runs = self.run_list()
        for run in runs:
            run.delete()
        self.df_app.delete()

    @classmethod
    def from_id(cls, id: str) -> "DataFlow":
        """
        Load a Data Flow job given an id.

        Parameters
        ----------
        id: str
            id of the Data Flow job to load

        Returns
        -------
        DataFlow
            a Data Flow job instance
        """
        payload = batch_convert_case(
            yaml.safe_load(repr(DataFlowApp.from_ocid(id))), "snake"
        )
        config = normalize_config(payload)
        name = config.pop("display_name")
        ins = cls(config)
        rt = DataFlowRuntime()
        for k, v in config.items():
            if hasattr(DataFlowRuntime, f"with_{k}"):
                getattr(rt, f"with_{k}")(v)
        ins.runtime = rt
        ins.name = name
        return ins

    @classmethod
    def list_jobs(cls, compartment_id: str = None, **kwargs) -> List["DataFlow"]:
        """
        List Data Flow jobs in a given compartment.

        Parameters
        ----------
        compartment_id: str
            id of that compartment
        kwargs
            additional keyword arguments for filtering jobs

        Returns
        -------
        List[DataFlow]
            list of Data Flow jobs
        """
        return [
            cls.from_id(job.id)
            for job in DataFlowApp.list_resource(compartment_id, **kwargs)
        ]

    def to_dict(self) -> dict:
        """
        Serialize job to a dictionary.

        Returns
        -------
        dict
            serialized job as a dictionary
        """
        spec = self._convert_shape_config(copy.deepcopy(self._spec), "camel")
        return {
            "kind": self.kind,
            "type": self.type,
            "spec": batch_convert_case(spec, "camel"),
        }

    @classmethod
    def from_dict(cls, config: dict) -> "DataFlow":
        """
        Load a Data Flow job instance from a dictionary of configurations.

        Parameters
        ----------
        config: dict
            dictionary of configurations

        Returns
        -------
        DataFlow
            a Data Flow job instance
        """
        spec = cls._convert_shape_config(copy.deepcopy(config["spec"]), "snake")
        return cls(spec=batch_convert_case(spec, "snake"))

    @class_or_instance_method
    def _convert_shape_config(cls, spec: Dict, to_format: str) -> Dict:
        """Converts the format of shape config details from camel to snake, or vice versa.

        Parameters
        ----------
        spec: dict
            dictionary of specs
        to_format: str
            the format that's converted to

        Returns
        -------
        Dict
            dictionary with converted shape config details
        """
        shape_config_map = [
            cls.CONST_DRIVER_SHAPE_CONFIG,
            cls.CONST_EXECUTOR_SHAPE_CONFIG,
            cls.attribute_map[cls.CONST_DRIVER_SHAPE_CONFIG],
            cls.attribute_map[cls.CONST_EXECUTOR_SHAPE_CONFIG],
        ]
        converted_map = {
            "camel": {
                cls.CONST_MEMORY_IN_GBS: cls.attribute_map[cls.CONST_MEMORY_IN_GBS],
                cls.CONST_OCPUS: cls.CONST_OCPUS,
            },
            "snake": {
                cls.attribute_map[cls.CONST_MEMORY_IN_GBS]: cls.CONST_MEMORY_IN_GBS,
                cls.CONST_OCPUS: cls.CONST_OCPUS,
            },
        }
        for shape_config in shape_config_map:
            shape_config_value = spec.pop(shape_config, {})
            if shape_config_value:
                temp_maps = {}
                for key, value in shape_config_value.items():
                    converted_key = converted_map[to_format].get(key, None)
                    if converted_key:
                        temp_maps[converted_key] = value
                    else:
                        temp_maps[key] = value
                spec[shape_config] = copy.deepcopy(temp_maps)
        return spec

    def to_yaml(self) -> str:
        """Serializes the object into YAML string.

        Returns
        -------
        str
            YAML stored in a string.
        """
        return yaml.safe_dump(self.to_dict())



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/from typing import Dict


from typing import Any, Dict
from ads.jobs.builders.base import Builder

MODEL_DEPLOYMENT_INFRASTRUCTURE_TYPE = "datascienceModelDeployment"
MODEL_DEPLOYMENT_INFRASTRUCTURE_KIND = "infrastructure"


class ModelDeploymentInfrastructure(Builder):
    """A class used to represent a Model Deployment Infrastructure.

    Attributes
    ----------
    compartment_id: str
        The compartment id of model deployment
    project_id: str
        The project id of model deployment
    shape_name: str
        The shape name of model deployment
    shape_config_details: Dict
        The shape config details of model deployment
    replica: int
        The replica of model deployment
    bandwidth_mbps: int
        The bandwidth of model deployment in mbps
    access_log: Dict
        The access log of model deployment
    predict_log: Dict
        The predict log of model deployment
    log_group_id: str
        The access and predict log group id of model deployment
    log_id: str
        The access and predict log id of model deployment
    web_concurrency: int
        The web concurrency of model deployment

    Methods
    -------
    with_compartment_id(compartment_id)
        Sets the compartment id of model deployment
    with_project_id(project_id)
        Sets the project id of model deployment
    with_shape_name(shape_name)
        Sets the shape name of model deployment
    with_shape_config_details(shape_config_details)
        Sets the shape config details of model deployment
    with_replica(replica)
        Sets the replica of model deployment
    with_bandwidth_mbps(bandwidth_mbps)
        Sets the bandwidth of model deployment in mbps
    with_access_log(log_group_id, log_id)
        Sets the access log of model deployment
    with_predict_log(log_group_id, log_id)
        Sets the predict log of model deployment
    with_log_group_id(log_group_id)
        Sets the access and predict log group id of model deployment
    with_log_id(log_id)
        Sets the access and predict log id of model deployment
    with_web_concurrency(web_concurrency)
        Sets the web concurrency of model deployment

    Example
    -------
    Build infrastructure from builder apis:
    >>> infrastructure = ModelDeploymentInfrastructure()
    ...        .with_project_id(<project_id>)
    ...        .with_compartment_id(<compartment_id>)
    ...        .with_shape_name("VM.Standard.E4.Flex")
    ...        .with_shape_config_details(
    ...            ocpus=1,
    ...            memory_in_gbs=16
    ...        )
    ...        .with_replica(1)
    ...        .with_bandwidth_mbps(10)
    ...        .with_web_concurrency(10)
    ...        .with_access_log(
    ...            log_group_id=<log_group_id>,
    ...            log_id=<log_id>
    ...        )
    ...        .with_predict_log(
    ...            log_group_id=<log_group_id>,
    ...            log_id=<log_id>
    ...        )
    >>> infrastructure.to_dict()

    Build infrastructure from yaml:
    >>> infrastructure = ModelDeploymentInfrastructure.from_yaml(uri=<path_to_yaml>)
    """

    CONST_PROJECT_ID = "projectId"
    CONST_COMPARTMENT_ID = "compartmentId"
    CONST_MODEL_DEPLOYMENT_CONFIG_DETAILS = "modelDeploymentConfigurationDetails"
    CONST_DEPLOYMENT_TYPE = "deploymentType"
    CONST_MODEL_CONFIG_DETAILS = "modelConfigurationDetails"
    CONST_INSTANCE_CONFIG = "instanceConfiguration"
    CONST_SHAPE_NAME = "shapeName"
    CONST_INSTANCE_SHAPE_NAME = "instanceShapeName"
    CONST_SHAPE_CONFIG_DETAILS = "shapeConfigDetails"
    CONST_MODEL_DEPLOYMENT_INSTANCE_SHAPE_CONFIG_DETAILS = (
        "modelDeploymentInstanceShapeConfigDetails"
    )
    CONST_OCPUS = "ocpus"
    CONST_MEMORY_IN_GBS = "memoryInGBs"
    CONST_SCALING_POLICY = "scalingPolicy"
    CONST_POLICY_TYPE = "policyType"
    CONST_REPLICA = "replica"
    CONST_INSTANCE_COUNT = "instanceCount"
    CONST_BANDWIDTH_MBPS = "bandwidthMbps"
    CONST_CATEGORY_LOG_DETAILS = "categoryLogDetails"
    CONST_ACCESS_LOG = "accessLog"
    CONST_ACCESS = "access"
    CONST_PREDICT_LOG = "predictLog"
    CONST_PREDICT = "predict"
    CONST_LOG_ID = "logId"
    CONST_LOG_GROUP_ID = "logGroupId"
    CONST_WEB_CONCURRENCY = "webConcurrency"
    CONST_STREAM_CONFIG_DETAILS = "streamConfigurationDetails"

    attribute_map = {
        CONST_PROJECT_ID: "project_id",
        CONST_COMPARTMENT_ID: "compartment_id",
        CONST_SHAPE_NAME: "shape_name",
        CONST_SHAPE_CONFIG_DETAILS: "shape_config_details",
        CONST_OCPUS: "ocpus",
        CONST_MEMORY_IN_GBS: "memory_in_gbs",
        CONST_REPLICA: "replica",
        CONST_BANDWIDTH_MBPS: "bandwidth_mbps",
        CONST_ACCESS_LOG: "access_log",
        CONST_PREDICT_LOG: "predict_log",
        CONST_LOG_ID: "log_id",
        CONST_LOG_GROUP_ID: "log_group_id",
        CONST_WEB_CONCURRENCY: "web_concurrency",
    }

    shape_config_details_attribute_map = {
        CONST_MEMORY_IN_GBS: "memory_in_gbs",
        CONST_OCPUS: "ocpus",
    }

    access_log_attribute_map = {
        CONST_LOG_ID: "log_id",
        CONST_LOG_GROUP_ID: "log_group_id",
    }

    predict_log_attribute_map = {
        CONST_LOG_ID: "log_id",
        CONST_LOG_GROUP_ID: "log_group_id",
    }

    MODEL_CONFIG_DETAILS_PATH = (
        "model_deployment_configuration_details.model_configuration_details"
    )

    payload_attribute_map = {
        CONST_PROJECT_ID: "project_id",
        CONST_COMPARTMENT_ID: "compartment_id",
        CONST_SHAPE_NAME: f"{MODEL_CONFIG_DETAILS_PATH}.instance_configuration.instance_shape_name",
        CONST_SHAPE_CONFIG_DETAILS: f"{MODEL_CONFIG_DETAILS_PATH}.instance_configuration.model_deployment_instance_shape_config_details",
        CONST_REPLICA: f"{MODEL_CONFIG_DETAILS_PATH}.scaling_policy.instance_count",
        CONST_BANDWIDTH_MBPS: f"{MODEL_CONFIG_DETAILS_PATH}.bandwidth_mbps",
        CONST_ACCESS_LOG: "category_log_details.access",
        CONST_PREDICT_LOG: "category_log_details.predict",
        CONST_DEPLOYMENT_TYPE: "model_deployment_configuration_details.deployment_type",
        CONST_POLICY_TYPE: f"{MODEL_CONFIG_DETAILS_PATH}.scaling_policy.policy_type",
    }

    sub_level_attribute_maps = {
        CONST_SHAPE_CONFIG_DETAILS: shape_config_details_attribute_map,
        CONST_ACCESS_LOG: access_log_attribute_map,
        CONST_PREDICT_LOG: predict_log_attribute_map,
    }

    def __init__(self, spec: Dict = None, **kwargs) -> None:
        super().__init__(spec, **kwargs)

    @property
    def kind(self) -> str:
        """The kind of the object as showing in YAML.

        Returns
        -------
        str
            infrastructure
        """
        return MODEL_DEPLOYMENT_INFRASTRUCTURE_KIND

    @property
    def type(self) -> str:
        """The type of the object as showing in YAML.

        Returns
        -------
        str
            datascienceModelDeployment
        """
        return MODEL_DEPLOYMENT_INFRASTRUCTURE_TYPE

    @property
    def compartment_id(self) -> str:
        """The model deployment compartment id.

        Returns
        -------
        str
            The model deployment compartment id.
        """
        return self.get_spec(self.CONST_COMPARTMENT_ID, None)

    def with_compartment_id(
        self, compartment_id: str
    ) -> "ModelDeploymentInfrastructure":
        """Sets the compartment id of model deployment.

        Parameters
        ----------
        compartment_id: str
            The compartment id of model deployment.

        Returns
        -------
        ModelDeploymentInfrastructure
            The ModelDeploymentInfrastructure instance (self).
        """
        return self.set_spec(self.CONST_COMPARTMENT_ID, compartment_id)

    @property
    def project_id(self) -> str:
        """The model deployment project id.

        Returns
        -------
        str
            The model deployment project id.
        """
        return self.get_spec(self.CONST_PROJECT_ID, None)

    def with_project_id(self, project_id: str) -> "ModelDeploymentInfrastructure":
        """Sets the project id of model deployment.

        Parameters
        ----------
        project_id: str
            The project id of model deployment.

        Returns
        -------
        ModelDeploymentInfrastructure
            The ModelDeploymentInfrastructure instance (self).
        """
        return self.set_spec(self.CONST_PROJECT_ID, project_id)

    @property
    def shape_config_details(self) -> Dict:
        """The model deployment shape config details.

        Returns
        -------
        Dict
            The model deployment shape config details.
        """
        return self.get_spec(self.CONST_SHAPE_CONFIG_DETAILS, {})

    def with_shape_config_details(
        self, ocpus: float, memory_in_gbs: float, **kwargs: Dict[str, Any]
    ) -> "ModelDeploymentInfrastructure":
        """Sets the shape config details of model deployment.

        Parameters
        ----------
        ocpus: float
            The ocpus number.
        memory_in_gbs: float
            The memory in gbs number.
        kwargs: Dict

        Returns
        -------
        ModelDeploymentInfrastructure
            The ModelDeploymentInfrastructure instance (self).
        """
        return self.set_spec(
            self.CONST_SHAPE_CONFIG_DETAILS,
            {
                self.CONST_OCPUS: ocpus,
                self.CONST_MEMORY_IN_GBS: memory_in_gbs,
                **kwargs,
            },
        )

    @property
    def shape_name(self) -> str:
        """The model deployment shape name.

        Returns
        -------
        str
            The model deployment shape name.
        """
        return self.get_spec(self.CONST_SHAPE_NAME, None)

    def with_shape_name(self, shape_name: str) -> "ModelDeploymentInfrastructure":
        """Sets the shape name of model deployment.

        Parameters
        ----------
        shape_name: str
            The shape name of model deployment.

        Returns
        -------
        ModelDeploymentInfrastructure
            The ModelDeploymentInfrastructure instance (self).
        """
        return self.set_spec(self.CONST_SHAPE_NAME, shape_name)

    @property
    def replica(self) -> int:
        """The model deployment instance count.

        Returns
        -------
        int
            The model deployment instance count.
        """
        return self.get_spec(self.CONST_REPLICA, None)

    def with_replica(self, replica: int) -> "ModelDeploymentInfrastructure":
        """Sets the instance count of model deployment.

        Parameters
        ----------
        replica: int
            The instance count of model deployment.

        Returns
        -------
        ModelDeploymentInfrastructure
            The ModelDeploymentInfrastructure instance (self).
        """
        return self.set_spec(self.CONST_REPLICA, replica)

    @property
    def bandwidth_mbps(self) -> int:
        """The model deployment bandwidth in mbps.

        Returns
        -------
        int
            The model deployment bandwidth in mbps.
        """
        return self.get_spec(self.CONST_BANDWIDTH_MBPS, None)

    def with_bandwidth_mbps(
        self, bandwidth_mbps: int
    ) -> "ModelDeploymentInfrastructure":
        """Sets the bandwidth of model deployment.

        Parameters
        ----------
        bandwidth_mbps: int
            The bandwidth of model deployment in mbps.

        Returns
        -------
        ModelDeploymentInfrastructure
            The ModelDeploymentInfrastructure instance (self).
        """
        return self.set_spec(self.CONST_BANDWIDTH_MBPS, bandwidth_mbps)

    @property
    def access_log(self) -> Dict:
        """The model deployment access log.

        Returns
        -------
        Dict
            The model deployment access log.
        """
        return self.get_spec(self.CONST_ACCESS_LOG, {})

    def with_access_log(
        self, log_group_id: str, log_id: str
    ) -> "ModelDeploymentInfrastructure":
        """Sets the access log of model deployment.

        Parameters
        ----------
        log_group_id: str
            The access log group id of model deployment.
        log_id: str
            The access log id of model deployment.

        Returns
        -------
        ModelDeploymentInfrastructure
            The ModelDeploymentInfrastructure instance (self).
        """
        return self.set_spec(
            self.CONST_ACCESS_LOG,
            {self.CONST_LOG_GROUP_ID: log_group_id, self.CONST_LOG_ID: log_id},
        )

    @property
    def predict_log(self) -> Dict:
        """The model deployment predict log.

        Returns
        -------
        Dict
            The model deployment predict log.
        """
        return self.get_spec(self.CONST_PREDICT_LOG, {})

    def with_predict_log(
        self, log_group_id: str, log_id: str
    ) -> "ModelDeploymentInfrastructure":
        """Sets the predict log of model deployment.

        Parameters
        ----------
        log_group_id: str
            The predict log group id of model deployment.
        log_id: str
            The predict log id of model deployment.

        Returns
        -------
        ModelDeploymentInfrastructure
            The ModelDeploymentInfrastructure instance (self).
        """
        return self.set_spec(
            self.CONST_PREDICT_LOG,
            {self.CONST_LOG_GROUP_ID: log_group_id, self.CONST_LOG_ID: log_id},
        )

    @property
    def log_group_id(self) -> str:
        """The model deployment log group id.

        Returns
        -------
        str
            The model deployment log group id.
        """
        return self.get_spec(self.CONST_LOG_GROUP_ID, None)

    def with_log_group_id(self, log_group_id: str) -> "ModelDeploymentInfrastructure":
        """Sets the log group id of model deployment.

        Parameters
        ----------
        log_group_id: str
            The predict and access log group id of model deployment.

        Returns
        -------
        ModelDeploymentInfrastructure
            The ModelDeploymentInfrastructure instance (self).
        """
        return self.set_spec(self.CONST_LOG_GROUP_ID, log_group_id)

    @property
    def log_id(self) -> str:
        """The model deployment log id.

        Returns
        -------
        str
            The model deployment log id.
        """
        return self.get_spec(self.CONST_LOG_ID, None)

    def with_log_id(self, log_id: str) -> "ModelDeploymentInfrastructure":
        """Sets the log id of model deployment.

        Parameters
        ----------
        log_id: str
            The predict and access log id of model deployment.

        Returns
        -------
        ModelDeploymentInfrastructure
            The ModelDeploymentInfrastructure instance (self).
        """
        return self.set_spec(self.CONST_LOG_ID, log_id)

    @property
    def web_concurrency(self) -> int:
        """The model deployment web concurrency.

        Returns
        -------
        int
            The model deployment web concurrency.
        """
        return self.get_spec(self.CONST_WEB_CONCURRENCY, None)

    def with_web_concurrency(
        self, web_concurrency: int
    ) -> "ModelDeploymentInfrastructure":
        """Sets the web concurrency of model deployment.

        Parameters
        ----------
        web_concurrency: int
            The web concurrency of model deployment.

        Returns
        -------
        ModelDeploymentInfrastructure
            The ModelDeploymentInfrastructure instance (self).
        """
        return self.set_spec(self.CONST_WEB_CONCURRENCY, web_concurrency)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from logging import getLogger
from collections import namedtuple
import fsspec
import yaml
import importlib

logger = getLogger("ads.yaml")


class YamlSpecParser:
    def translate_config(self, config: list):
        """
        Translates config to three element tuple of dictionary- (file, envVars, concatenated args)

        All elements that are of type `env` goes to envVars
        All elements that are of type `args` are concatenated as `--key1=value1 --key2=value2 ..` string
        All elements of type `file` are currently ignored
        """
        # TODO parse file type

        parsed_config = namedtuple(
            "ClusterConfig", field_names=["files", "envVars", "cmd_args"]
        )
        envVars = {}
        files = {}
        cmd_args = ""
        if config:
            if config.get("env"):
                for item in config["env"]:
                    envVars[item["name"]] = item["value"]
            cmd_args = " ".join(config.get("startOptions", [""]))
        return parsed_config(files=files, envVars=envVars, cmd_args=cmd_args)

    def parse(self):
        pass

    @classmethod
    def parse_content(cls, file):
        logger.debug(f"Config filename is {file}")
        yaml_spec = {}
        if isinstance(file, dict):
            yaml_spec = file
        elif isinstance(file, str):
            with fsspec.open(file) as yf:
                yaml_spec = yaml.load(yf.read(), yaml.SafeLoader)
        kind = yaml_spec.get("kind")
        parsed_output = None
        if kind:
            className = f"{kind[0].upper()}{kind[1:]}SpecParser"
            logger.debug(f"kind: {kind}")
            logger.debug(f"Parser Class name: {className}")
            m = importlib.import_module("ads.opctl.config.yaml_parsers")
            parser = getattr(m, className)
            parsed_output = parser(yaml_spec).parse()
            logger.debug(f"Parsed Output: {parsed_output}")
        return parsed_output



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


import collections
import copy
import datetime
import oci
import warnings
import time
from typing import Dict, List, Union, Any

import oci.loggingsearch
from ads.common import auth as authutil
import pandas as pd
from ads.model.serde.model_input import JsonModelInputSERDE
from ads.common import auth, oci_client
from ads.common.oci_logging import (
    LOG_INTERVAL,
    LOG_RECORDS_LIMIT,
    ConsolidatedLog,
    OCILog,
)
from ads.config import COMPARTMENT_OCID, PROJECT_OCID
from ads.jobs.builders.base import Builder
from ads.jobs.builders.infrastructure.utils import get_value
from ads.model.common.utils import _is_json_serializable
from ads.model.deployment.common.utils import send_request
from ads.model.deployment.model_deployment_infrastructure import (
    MODEL_DEPLOYMENT_INFRASTRUCTURE_TYPE,
    ModelDeploymentInfrastructure,
)
from ads.model.deployment.model_deployment_runtime import (
    ModelDeploymentCondaRuntime,
    ModelDeploymentContainerRuntime,
    ModelDeploymentRuntime,
    ModelDeploymentRuntimeType,
    OCIModelDeploymentRuntimeType,
)
from ads.model.service.oci_datascience_model_deployment import (
    OCIDataScienceModelDeployment,
)
from ads.common import utils as ads_utils
from .common import utils
from .common.utils import OCIClientManager, State
from .model_deployment_properties import ModelDeploymentProperties
from oci.data_science.models import (
    LogDetails,
    CreateModelDeploymentDetails,
    UpdateModelDeploymentDetails,
)

DEFAULT_WAIT_TIME = 1200
DEFAULT_POLL_INTERVAL = 10
DEFAULT_WORKFLOW_STEPS = 6
DELETE_WORKFLOW_STEPS = 2
DEACTIVATE_WORKFLOW_STEPS = 2
DEFAULT_RETRYING_REQUEST_ATTEMPTS = 3
TERMINAL_STATES = [State.ACTIVE, State.FAILED, State.DELETED, State.INACTIVE]

MODEL_DEPLOYMENT_KIND = "deployment"
MODEL_DEPLOYMENT_TYPE = "modelDeployment"


class ModelDeploymentLogType:
    PREDICT = "predict"
    ACCESS = "access"


class ModelDeploymentMode:
    HTTPS = "HTTPS_ONLY"
    STREAM = "STREAM_ONLY"


class LogNotConfiguredError(Exception):
    pass


class ModelDeploymentFailedError(Exception):
    pass


class ModelDeployment(Builder):
    """
    A class used to represent a Model Deployment.

    Attributes
    ----------
    config: (dict)
        Deployment configuration parameters
    properties: (ModelDeploymentProperties)
        ModelDeploymentProperties object
    workflow_state_progress: (str)
        Workflow request id
    workflow_steps: (int)
        The number of steps in the workflow
    dsc_model_deployment: (OCIDataScienceModelDeployment)
        The OCIDataScienceModelDeployment instance.
    state: (State)
        Returns the deployment state of the current Model Deployment object
    created_by: (str)
        The user that creates the model deployment
    lifecycle_state: (str)
        Model deployment lifecycle state
    lifecycle_details: (str)
        Model deployment lifecycle details
    time_created: (datetime)
        The time when the model deployment is created
    display_name: (str)
        Model deployment display name
    description: (str)
        Model deployment description
    freeform_tags: (dict)
        Model deployment freeform tags
    defined_tags: (dict)
        Model deployment defined tags
    runtime: (ModelDeploymentRuntime)
        Model deployment runtime
    infrastructure: (ModelDeploymentInfrastructure)
        Model deployment infrastructure


    Methods
    -------
    deploy(wait_for_completion, **kwargs)
        Deploy the current Model Deployment object
    delete(wait_for_completion, **kwargs)
        Deletes the current Model Deployment object
    update(wait_for_completion, **kwargs)
        Updates a model deployment
    activate(wait_for_completion, max_wait_time, poll_interval)
        Activates a model deployment
    deactivate(wait_for_completion, max_wait_time, poll_interval)
        Deactivates a model deployment
    list(status, compartment_id, project_id, **kwargs)
        List model deployment within given compartment and project.
    with_display_name(display_name)
        Sets model deployment display name
    with_description(description)
        Sets model deployment description
    with_freeform_tags(freeform_tags)
        Sets model deployment freeform tags
    with_defined_tags(defined_tags)
        Sets model deployment defined tags
    with_runtime(self, runtime)
        Sets model deployment runtime
    with_infrastructure(self, infrastructure)
        Sets model deployment infrastructure
    from_dict(obj_dict)
        Deserializes model deployment instance from dict
    from_id(id)
        Loads model deployment instance from ocid
    sync()
        Updates the model deployment instance from backend


    Examples
    --------
    Build model deployment from builder apis:
    >>> ds_model_deployment = (ModelDeployment()
    ...    .with_display_name("TestModelDeployment")
    ...    .with_description("Testing the test model deployment")
    ...    .with_freeform_tags(tag1="val1", tag2="val2")
    ...    .with_infrastructure(
    ...        (ModelDeploymentInfrastructure()
    ...        .with_project_id(<project_id>)
    ...        .with_compartment_id(<compartment_id>)
    ...        .with_shape_name("VM.Standard.E4.Flex")
    ...        .with_shape_config_details(
    ...            ocpus=1,
    ...            memory_in_gbs=16
    ...        )
    ...        .with_replica(1)
    ...        .with_bandwidth_mbps(10)
    ...        .with_web_concurrency(10)
    ...        .with_access_log(
    ...            log_group_id=<log_group_id>,
    ...            log_id=<log_id>
    ...        )
    ...        .with_predict_log(
    ...            log_group_id=<log_group_id>,
    ...            log_id=<log_id>
    ...        ))
    ...    )
    ...    .with_runtime(
    ...        (ModelDeploymentContainerRuntime()
    ...        .with_image(<image>)
    ...        .with_image_digest(<image_digest>)
    ...        .with_entrypoint(<entrypoint>)
    ...        .with_server_port(<server_port>)
    ...        .with_health_check_port(<health_check_port>)
    ...        .with_env({"key":"value"})
    ...        .with_deployment_mode("HTTPS_ONLY")
    ...        .with_model_uri(<model_uri>))
    ...    )
    ... )
    >>> ds_model_deployment.deploy()
    >>> ds_model_deployment.status
    >>> ds_model_deployment.with_display_name("new name").update()
    >>> ds_model_deployment.deactivate()
    >>> ds_model_deployment.sync()
    >>> ds_model_deployment.list(status="ACTIVE")
    >>> ds_model_deployment.delete()

    Build model deployment from yaml
    >>> ds_model_deployment = ModelDeployment.from_yaml(uri=<path_to_yaml>)
    """

    _PREFIX = "datascience_model_deployment"

    CONST_ID = "id"
    CONST_CREATED_BY = "createdBy"
    CONST_DISPLAY_NAME = "displayName"
    CONST_DESCRIPTION = "description"
    CONST_FREEFORM_TAG = "freeformTags"
    CONST_DEFINED_TAG = "definedTags"
    CONST_MODEL_DEPLOYMENT_URL = "modelDeploymentUrl"
    CONST_INFRASTRUCTURE = "infrastructure"
    CONST_RUNTIME = "runtime"
    CONST_LIFECYCLE_STATE = "lifecycleState"
    CONST_LIFECYCLE_DETAILS = "lifecycleDetails"
    CONST_TIME_CREATED = "timeCreated"

    attribute_map = {
        CONST_ID: "id",
        CONST_CREATED_BY: "created_by",
        CONST_DISPLAY_NAME: "display_name",
        CONST_DESCRIPTION: "description",
        CONST_FREEFORM_TAG: "freeform_tags",
        CONST_DEFINED_TAG: "defined_tags",
        CONST_MODEL_DEPLOYMENT_URL: "model_deployment_url",
        CONST_INFRASTRUCTURE: "infrastructure",
        CONST_RUNTIME: "runtime",
        CONST_LIFECYCLE_STATE: "lifecycle_state",
        CONST_LIFECYCLE_DETAILS: "lifecycle_details",
        CONST_TIME_CREATED: "time_created",
    }

    initialize_spec_attributes = [
        "display_name",
        "description",
        "freeform_tags",
        "defined_tags",
        "infrastructure",
        "runtime",
    ]
    model_input_serializer = JsonModelInputSERDE()

    def __init__(
        self,
        properties: Union[ModelDeploymentProperties, Dict] = None,
        config: Dict = None,
        model_deployment_id: str = None,
        model_deployment_url: str = "",
        spec: Dict = None,
        **kwargs,
    ):
        """Initializes a ModelDeployment object.

        Parameters
        ----------
        properties: (Union[ModelDeploymentProperties, Dict], optional). Defaults to None.
            Object containing deployment properties.
            The properties can be `None` when `kwargs` are used for specifying properties.
        config: (Dict, optional). Defaults to None.
            ADS auth dictionary for OCI authentication.
            This can be generated by calling `ads.common.auth.api_keys()` or `ads.common.auth.resource_principal()`.
            If this is `None` then the `ads.common.default_signer(client_kwargs)` will be used.
        model_deployment_id: (str, optional). Defaults to None.
            Model deployment OCID.
        model_deployment_url: (str, optional). Defaults to empty string.
            Model deployment url.
        spec: (dict, optional). Defaults to None.
            Model deployment spec.
        kwargs:
            Keyword arguments for initializing `ModelDeploymentProperties` or `ModelDeployment`.
        """

        if spec and properties:
            raise ValueError(
                "You can only pass in either `spec` or `properties` to initialize model deployment instance."
            )

        if config:
            warnings.warn(
                "`config` will be deprecated in 3.0.0 and will be ignored now. Please use `ads.set_auth()` to config the auth information."
            )

        if properties:
            warnings.warn(
                "`properties` will be deprecated in 3.0.0. Please use `spec` or the builder pattern to initialize model deployment instance."
            )

        if model_deployment_url or model_deployment_id:
            warnings.warn(
                "`model_deployment_url` and `model_deployment_id` will be deprecated in 3.0.0 and will be ignored now. These two fields will be auto-populated from the service side."
            )

        initialize_spec = {}
        initialize_spec_kwargs = {}
        if spec:
            initialize_spec = spec
            initialize_spec_kwargs = self._extract_spec_kwargs(**kwargs)
        elif not properties and not spec:
            if self.CONST_INFRASTRUCTURE in kwargs or self.CONST_RUNTIME in kwargs:
                initialize_spec_kwargs = self._extract_spec_kwargs(**kwargs)

        super().__init__(spec=initialize_spec, **initialize_spec_kwargs)

        self.properties = (
            properties
            if isinstance(properties, ModelDeploymentProperties)
            else ModelDeploymentProperties(oci_model_deployment=properties, **kwargs)
        )

        self.current_state = (
            State._from_str(self.properties.lifecycle_state)
            if self.properties.lifecycle_state
            else State.UNKNOWN
        )

        self._access_log = None
        self._predict_log = None
        self.dsc_model_deployment = OCIDataScienceModelDeployment()

    @property
    def kind(self) -> str:
        """The kind of the object as showing in YAML.

        Returns
        -------
        str
            deployment
        """
        return MODEL_DEPLOYMENT_KIND

    @property
    def type(self) -> str:
        """The type of the object as showing in YAML.

        Returns
        -------
        str
            deployment
        """
        return MODEL_DEPLOYMENT_TYPE

    @property
    def model_deployment_id(self) -> str:
        """The model deployment ocid.

        Returns
        -------
        str
            The model deployment ocid.
        """
        return self.get_spec(self.CONST_ID, None)

    @property
    def created_by(self) -> str:
        """The user that creates the model deployment.

        Returns
        -------
        str
            The user that creates the model deployment.
        """
        return self.get_spec(self.CONST_CREATED_BY, None)

    @property
    def url(self) -> str:
        """Model deployment url.

        Returns
        -------
        str
            Model deployment url.
        """
        return self.get_spec(self.CONST_MODEL_DEPLOYMENT_URL, None)

    @property
    def lifecycle_state(self) -> str:
        """Model deployment lifecycle state.

        Returns
        -------
        str
            Model deployment lifecycle state.
        """
        return self.get_spec(self.CONST_LIFECYCLE_STATE, None)

    @property
    def lifecycle_details(self) -> str:
        """Model deployment lifecycle details.

        Returns
        -------
        str
            Model deployment lifecycle details.
        """
        return self.get_spec(self.CONST_LIFECYCLE_DETAILS, None)

    @property
    def time_created(self) -> datetime:
        """The time when the model deployment is created.

        Returns
        -------
        datetime
            The time when the model deployment is created.
        """
        return self.get_spec(self.CONST_TIME_CREATED, None)

    @property
    def display_name(self) -> str:
        """Model deployment display name.

        Returns
        -------
        str
            Model deployment display name.
        """
        return self.get_spec(self.CONST_DISPLAY_NAME, None)

    def with_display_name(self, display_name: str) -> "ModelDeployment":
        """Sets the name of model deployment.

        Parameters
        ----------
        display_name: str
            The name of model deployment.

        Returns
        -------
        ModelDeployment
            The ModelDeployment instance (self).
        """
        return self.set_spec(self.CONST_DISPLAY_NAME, display_name)

    @property
    def description(self) -> str:
        """Model deployment description.

        Returns
        -------
        str
            Model deployment description.
        """
        return self.get_spec(self.CONST_DESCRIPTION, None)

    def with_description(self, description: str) -> "ModelDeployment":
        """Sets the description of model deployment.

        Parameters
        ----------
        description: str
            The description of model deployment.

        Returns
        -------
        ModelDeployment
            The ModelDeployment instance (self).
        """
        return self.set_spec(self.CONST_DESCRIPTION, description)

    @property
    def freeform_tags(self) -> Dict:
        """Model deployment freeform tags.

        Returns
        -------
        Dict
            Model deployment freeform tags.
        """
        return self.get_spec(self.CONST_FREEFORM_TAG, {})

    def with_freeform_tags(self, **kwargs) -> "ModelDeployment":
        """Sets the freeform tags of model deployment.

        Parameters
        ----------
        kwargs
            The freeform tags of model deployment.

        Returns
        -------
        ModelDeployment
            The ModelDeployment instance (self).
        """
        return self.set_spec(self.CONST_FREEFORM_TAG, kwargs)

    @property
    def defined_tags(self) -> Dict:
        """Model deployment defined tags.

        Returns
        -------
        Dict
            Model deployment defined tags.
        """
        return self.get_spec(self.CONST_DEFINED_TAG, {})

    def with_defined_tags(self, **kwargs) -> "ModelDeployment":
        """Sets the defined tags of model deployment.

        Parameters
        ----------
        kwargs
            The defined tags of model deployment.

        Returns
        -------
        ModelDeployment
            The ModelDeployment instance (self).
        """
        return self.set_spec(self.CONST_DEFINED_TAG, kwargs)

    @property
    def runtime(self) -> "ModelDeploymentRuntime":
        """Model deployment runtime.

        Returns
        -------
        ModelDeploymentRuntime
            Model deployment runtime.
        """
        return self.get_spec(self.CONST_RUNTIME, None)

    def with_runtime(self, runtime: ModelDeploymentRuntime) -> "ModelDeployment":
        """Sets the runtime of model deployment.

        Parameters
        ----------
        runtime: ModelDeploymentRuntime
            The runtime of model deployment.

        Returns
        -------
        ModelDeployment
            The ModelDeployment instance (self).
        """
        return self.set_spec(self.CONST_RUNTIME, runtime)

    @property
    def infrastructure(self) -> "ModelDeploymentInfrastructure":
        """Model deployment infrastructure.

        Returns
        -------
        ModelDeploymentInfrastructure
            Model deployment infrastructure.
        """
        return self.get_spec(self.CONST_INFRASTRUCTURE, None)

    def with_infrastructure(
        self, infrastructure: ModelDeploymentInfrastructure
    ) -> "ModelDeployment":
        """Sets the infrastructure of model deployment.

        Parameters
        ----------
        infrastructure: ModelDeploymentInfrastructure
            The infrastructure of model deployment.

        Returns
        -------
        ModelDeployment
            The ModelDeployment instance (self).
        """
        return self.set_spec(self.CONST_INFRASTRUCTURE, infrastructure)

    def deploy(
        self,
        wait_for_completion: bool = True,
        max_wait_time: int = DEFAULT_WAIT_TIME,
        poll_interval: int = DEFAULT_POLL_INTERVAL,
    ):
        """Deploys the current ModelDeployment object

        Parameters
        ----------
        wait_for_completion: bool
            Flag set for whether to wait for deployment to be deployed before proceeding.
            Defaults to True.
        max_wait_time: int
            Maximum amount of time to wait in seconds (Defaults to 1200).
            Negative implies infinite wait time.
        poll_interval: int
            Poll interval in seconds (Defaults to 10).

        Returns
        -------
        ModelDeployment
           The instance of ModelDeployment.

        Raises
        ------
        ModelDeploymentFailedError
            If model deployment fails to deploy
        """
        create_model_deployment_details = (
            self._build_model_deployment_details()
            if self._spec
            else self.properties.build()
        )

        response = self.dsc_model_deployment.create(
            create_model_deployment_details=create_model_deployment_details,
            wait_for_completion=wait_for_completion,
            max_wait_time=max_wait_time,
            poll_interval=poll_interval,
        )

        if response.lifecycle_state == State.FAILED.name:
            raise ModelDeploymentFailedError(
                f"Model deployment {response.id} failed to deploy: {response.lifecycle_details}"
            )

        return self._update_from_oci_model(response)

    def delete(
        self,
        wait_for_completion: bool = True,
        max_wait_time: int = DEFAULT_WAIT_TIME,
        poll_interval: int = DEFAULT_POLL_INTERVAL,
    ):
        """Deletes the ModelDeployment

        Parameters
        ----------
        wait_for_completion: bool
            Flag set for whether to wait for deployment to be deleted before proceeding.
            Defaults to True.
        max_wait_time: int
            Maximum amount of time to wait in seconds (Defaults to 1200).
            Negative implies infinite wait time.
        poll_interval: int
            Poll interval in seconds (Defaults to 10).

        Returns
        -------
        ModelDeployment
            The instance of ModelDeployment.
        """
        response = self.dsc_model_deployment.delete(
            wait_for_completion=wait_for_completion,
            max_wait_time=max_wait_time,
            poll_interval=poll_interval,
        )
        return self._update_from_oci_model(response)

    def update(
        self,
        properties: Union[ModelDeploymentProperties, dict, None] = None,
        wait_for_completion: bool = True,
        max_wait_time: int = DEFAULT_WAIT_TIME,
        poll_interval: int = DEFAULT_POLL_INTERVAL,
        **kwargs,
    ):
        """Updates a model deployment

        You can update `model_deployment_configuration_details` and change `instance_shape` and `model_id`
        when the model deployment is in the ACTIVE lifecycle state.
        The `bandwidth_mbps` or `instance_count` can only be updated while the model deployment is in the `INACTIVE` state.
        Changes to the `bandwidth_mbps` or `instance_count` will take effect the next time
        the `ActivateModelDeployment` action is invoked on the model deployment resource.

        Parameters
        ----------
        properties: ModelDeploymentProperties or dict
            The properties for updating the deployment.
        wait_for_completion: bool
            Flag set for whether to wait for deployment to be updated before proceeding.
            Defaults to True.
        max_wait_time: int
            Maximum amount of time to wait in seconds (Defaults to 1200).
            Negative implies infinite wait time.
        poll_interval: int
            Poll interval in seconds (Defaults to 10).
        kwargs:
            dict

        Returns
        -------
        ModelDeployment
            The instance of ModelDeployment.
        """
        updated_properties = properties
        if not isinstance(properties, ModelDeploymentProperties):
            updated_properties = ModelDeploymentProperties(
                oci_model_deployment=properties, **kwargs
            )

        update_model_deployment_details = (
            updated_properties.to_update_deployment()
            if properties or updated_properties.oci_model_deployment or kwargs
            else self._update_model_deployment_details()
        )

        response = self.dsc_model_deployment.update(
            update_model_deployment_details=update_model_deployment_details,
            wait_for_completion=wait_for_completion,
            max_wait_time=max_wait_time,
            poll_interval=poll_interval,
        )

        return self._update_from_oci_model(response)

    def watch(
        self,
        log_type: str = ModelDeploymentLogType.ACCESS,
        time_start: datetime = None,
        interval: int = LOG_INTERVAL,
        log_filter: str = None,
    ) -> "ModelDeployment":
        """Streams the access and/or predict logs of model deployment.

        Parameters
        ----------
        log_type: str, optional
            The log type. Can be `access`, `predict` or None.
            Defaults to access.
        time_start : datetime.datetime, optional
            Starting time for the log query.
            Defaults to None.
        interval : int, optional
            The time interval between sending each request to pull logs from OCI logging service.
            Defaults to 3.
        log_filter : str, optional
            Expression for filtering the logs. This will be the WHERE clause of the query.
            Defaults to None.

        Returns
        -------
        ModelDeployment
            The instance of ModelDeployment.
        """
        status = ""
        while not self._stop_condition():
            status = self._check_and_print_status(status)
            time.sleep(interval)

        time_start = time_start or self.time_created
        try:
            count = self.logs(log_type).stream(
                source=self.model_deployment_id,
                interval=interval,
                stop_condition=self._stop_condition,
                time_start=time_start,
                log_filter=log_filter,
            )

            if not count:
                print(
                    "No logs in the last 14 days. Please reset time_start to see older logs."
                )
            return self.sync()
        except KeyboardInterrupt:
            print("Stop watching logs.")
            pass

    def _stop_condition(self):
        """Stops the sync once the model deployment is in a terminal state."""
        return self.state in TERMINAL_STATES

    def _check_and_print_status(self, prev_status) -> str:
        """Check and print the next status.

        Parameters
        ----------
        prev_status: str
            The previous model deployment status.

        Returns
        -------
        str:
            The next model deployment status.
        """
        status = self._model_deployment_status_text()
        if status != prev_status:
            timestamp = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
            print(f"{timestamp} - {status}")
        return status

    def _model_deployment_status_text(self) -> str:
        """Formats the status message.

        Returns
        -------
        str:
            The model deployment life status and life cycle details.
        """
        details = f", {self.lifecycle_details}" if self.lifecycle_details else ""
        return f"Model Deployment {self.lifecycle_state}" + details

    @property
    def state(self) -> State:
        """Returns the deployment state of the current Model Deployment object"""
        request_attempts = 0
        while request_attempts < DEFAULT_RETRYING_REQUEST_ATTEMPTS:
            request_attempts += 1
            try:
                self.current_state = State._from_str(self.sync().lifecycle_state)
                break
            except:
                pass
            time.sleep(1)

        return self.current_state

    @property
    def status(self) -> State:
        """Returns the deployment state of the current Model Deployment object"""
        return self.state

    def predict(
        self,
        json_input=None,
        data: Any = None,
        serializer: "ads.model.ModelInputSerializer" = model_input_serializer,
        auto_serialize_data: bool = False,
        **kwargs,
    ) -> dict:
        """Returns prediction of input data run against the model deployment endpoint.

        Examples
        --------
        >>> import numpy as np
        >>> from ads.model import ModelInputSerializer
        >>> class MySerializer(ModelInputSerializer):
        ...     def serialize(self, data):
        ...         serialized_data = 1
        ...         return serialized_data
        >>> model_deployment = ModelDeployment.from_id(<model_deployment_id>)
        >>> prediction = model_deployment.predict(
        ...        data=np.array([1, 2, 3]),
        ...        serializer=MySerializer(),
        ...        auto_serialize_data=True,
        ... )['prediction']

        Parameters
        ----------
        json_input: Json serializable
            JSON payload for the prediction.
        data: Any
            Data for the prediction.
        serializer: ads.model.ModelInputSerializer
            Defaults to ads.model.JsonModelInputSerializer.
        auto_serialize_data: bool
            Defaults to False. Indicate whether to auto serialize input data using `serializer`.
            If `auto_serialize_data=False`, `data` required to be bytes or json serializable
            and `json_input` required to be json serializable. If `auto_serialize_data` set
            to True, data will be serialized before sending to model deployment endpoint.
        kwargs:
            content_type: str
                Used to indicate the media type of the resource.
                By default, it will be `application/octet-stream` for bytes input and `application/json` otherwise.
                The content-type header will be set to this value when calling the model deployment endpoint.

        Returns
        -------
        dict:
            Prediction results.

        """
        endpoint = f"{self.url}/predict"
        signer = authutil.default_signer()["signer"]
        header = {
            "signer": signer,
            "content_type": kwargs.get("content_type", None),
        }

        if data is None and json_input is None:
            raise AttributeError(
                "Neither `data` nor `json_input` are provided. You need to provide one of them."
            )
        if data is not None and json_input is not None:
            raise AttributeError(
                "`data` and `json_input` are both provided. You can only use one of them."
            )

        if auto_serialize_data:
            data = data or json_input
            serialized_data = serializer.serialize(data=data)
            return send_request(
                data=serialized_data,
                endpoint=endpoint,
                is_json_payload=_is_json_serializable(serialized_data),
                header=header,
            )

        if json_input is not None:
            if not _is_json_serializable(json_input):
                raise ValueError(
                    "`json_input` must be json serializable. "
                    "Set `auto_serialize_data` to True, or serialize the provided input data first,"
                    "or using `data` to pass binary data."
                )
            utils.get_logger().warning(
                "The `json_input` argument of `predict()` will be deprecated soon. "
                "Please use `data` argument. "
            )
            data = json_input

        is_json_payload = _is_json_serializable(data)
        if not isinstance(data, bytes) and not is_json_payload:
            raise TypeError(
                "`data` is not bytes or json serializable. Set `auto_serialize_data` to `True` to serialize the input data."
            )

        prediction = send_request(
            data=data, endpoint=endpoint, is_json_payload=is_json_payload, header=header
        )
        return prediction

    def activate(
        self,
        wait_for_completion: bool = True,
        max_wait_time: int = DEFAULT_WAIT_TIME,
        poll_interval: int = DEFAULT_POLL_INTERVAL,
    ) -> "ModelDeployment":
        """Activates a model deployment

        Parameters
        ----------
        wait_for_completion: bool
            Flag set for whether to wait for deployment to be activated before proceeding.
            Defaults to True.
        max_wait_time: int
            Maximum amount of time to wait in seconds (Defaults to 1200).
            Negative implies infinite wait time.
        poll_interval: int
            Poll interval in seconds (Defaults to 10).

        Returns
        -------
        ModelDeployment
            The instance of ModelDeployment.
        """
        response = self.dsc_model_deployment.activate(
            wait_for_completion=wait_for_completion,
            max_wait_time=max_wait_time,
            poll_interval=poll_interval,
        )

        return self._update_from_oci_model(response)

    def deactivate(
        self,
        wait_for_completion: bool = True,
        max_wait_time: int = DEFAULT_WAIT_TIME,
        poll_interval: int = DEFAULT_POLL_INTERVAL,
    ) -> "ModelDeployment":
        """Deactivates a model deployment

        Parameters
        ----------
        wait_for_completion: bool
            Flag set for whether to wait for deployment to be deactivated before proceeding.
            Defaults to True.
        max_wait_time: int
            Maximum amount of time to wait in seconds (Defaults to 1200).
            Negative implies infinite wait time.
        poll_interval: int
            Poll interval in seconds (Defaults to 10).

        Returns
        -------
        ModelDeployment
            The instance of ModelDeployment.
        """
        response = self.dsc_model_deployment.deactivate(
            wait_for_completion=wait_for_completion,
            max_wait_time=max_wait_time,
            poll_interval=poll_interval,
        )

        return self._update_from_oci_model(response)

    def _log_details(self, log_type: str = ModelDeploymentLogType.ACCESS):
        """Gets log details for the provided `log_type`.

        Properties
        ----------
        log_type: (str, optional). Defaults to "access".
            The log type. Can be "access" or "predict".

        Returns
        -------
        oci.datascience_model.CategoryLogDetails
            Category log details of the ModelDeployment.

        Raises
        ------
        AttributeError
            Deployment doesn't have requested log configuration.

        """
        if self.properties.category_log_details and getattr(
            self.properties.category_log_details, log_type
        ):
            return getattr(self.properties.category_log_details, log_type)
        elif self.infrastructure:
            category_log_details = self._build_category_log_details()
            log = category_log_details.get(log_type, None)
            if log and log.get("logId", None) and log.get("logGroupId", None):
                return LogDetails(
                    log_id=log.get("logId"), log_group_id=log.get("logGroupId")
                )

        raise LogNotConfiguredError(
            f"Deployment `{self.model_deployment_id}` "
            f"has no `{log_type}` log configuration."
        )

    @property
    def predict_log(self) -> OCILog:
        """Gets the model deployment predict logs object.

        Returns
        -------
        OCILog
            The OCILog object containing the predict logs.
        """
        if not self._predict_log:
            log_details = self._log_details(log_type=ModelDeploymentLogType.PREDICT)
            compartment_id = (
                self.infrastructure.compartment_id
                if self.infrastructure
                else self.properties.compartment_id
            )
            self._predict_log = OCILog(
                compartment_id=compartment_id or COMPARTMENT_OCID,
                id=log_details.log_id,
                log_group_id=log_details.log_group_id,
                source=self.model_deployment_id,
                annotation=ModelDeploymentLogType.PREDICT,
            )
        return self._predict_log

    @property
    def access_log(self) -> OCILog:
        """Gets the model deployment access logs object.

        Returns
        -------
        OCILog
            The OCILog object containing the access logs.
        """
        if not self._access_log:
            log_details = self._log_details(log_type=ModelDeploymentLogType.ACCESS)
            compartment_id = (
                self.infrastructure.compartment_id
                if self.infrastructure
                else self.properties.compartment_id
            )
            self._access_log = OCILog(
                compartment_id=compartment_id or COMPARTMENT_OCID,
                id=log_details.log_id,
                log_group_id=log_details.log_group_id,
                source=self.model_deployment_id,
                annotation=ModelDeploymentLogType.ACCESS,
            )
        return self._access_log

    def logs(self, log_type: str = None) -> ConsolidatedLog:
        """Gets the access or predict logs.

        Parameters
        ----------
        log_type: (str, optional). Defaults to None.
            The log type. Can be "access", "predict" or None.

        Returns
        -------
        ConsolidatedLog
            The ConsolidatedLog object containing the logs.
        """
        loggers = []
        if not log_type:
            try:
                loggers.append(self.access_log)
            except LogNotConfiguredError:
                pass

            try:
                loggers.append(self.predict_log)
            except LogNotConfiguredError:
                pass

            if not loggers:
                raise LogNotConfiguredError(
                    "Neither `predict` nor `access` log was configured for the model deployment."
                )
        elif log_type == ModelDeploymentLogType.ACCESS:
            loggers = [self.access_log]
        elif log_type == ModelDeploymentLogType.PREDICT:
            loggers = [self.predict_log]
        else:
            raise ValueError(
                "Parameter log_type should be either access, predict or None."
            )

        return ConsolidatedLog(*loggers)

    def show_logs(
        self,
        time_start: datetime.datetime = None,
        time_end: datetime.datetime = None,
        limit: int = LOG_RECORDS_LIMIT,
        log_type: str = None,
    ):
        """Shows deployment logs as a pandas dataframe.

        Parameters
        ----------
        time_start: (datetime.datetime, optional). Defaults to None.
            Starting date and time in RFC3339 format for retrieving logs.
            Defaults to None. Logs will be retrieved 14 days from now.
        time_end: (datetime.datetime, optional). Defaults to None.
            Ending date and time in RFC3339 format for retrieving logs.
            Defaults to None. Logs will be retrieved until now.
        limit: (int, optional). Defaults to 100.
            The maximum number of items to return.
        log_type: (str, optional). Defaults to None.
            The log type. Can be "access", "predict" or None.

        Returns
        -------
            A pandas DataFrame containing logs.
        """
        logging = self.logs(log_type=log_type)

        def prepare_log_record(log):
            """Converts a log record to ordered dict"""
            log_content = log.get("logContent", {})
            return collections.OrderedDict(
                [
                    ("type", log_content.get("type").split(".")[-1]),
                    ("id", log_content.get("id")),
                    ("message", log_content.get("data", {}).get("message")),
                    ("time", log_content.get("time")),
                ]
            )

        logs = logging.search(
            source=self.model_deployment_id,
            time_start=time_start,
            time_end=time_end,
            limit=limit,
        )
        return pd.DataFrame([prepare_log_record(log.data) for log in logs])

    def sync(self) -> "ModelDeployment":
        """Updates the model deployment instance from backend.

        Returns
        -------
        ModelDeployment
            The ModelDeployment instance (self).
        """
        return self._update_from_oci_model(
            OCIDataScienceModelDeployment.from_id(self.model_deployment_id)
        )

    @classmethod
    def list(
        cls,
        status: str = None,
        compartment_id: str = None,
        project_id: str = None,
        **kwargs,
    ) -> List["ModelDeployment"]:
        """Lists the model deployments associated with current compartment id and status

        Parameters
        ----------
        status : str
            Status of deployment. Defaults to None.
            Allowed values: `ACTIVE`, `CREATING`, `DELETED`, `DELETING`, `FAILED`, `INACTIVE` and `UPDATING`.
        compartment_id : str
            Target compartment to list deployments from.
            Defaults to the compartment set in the environment variable "NB_SESSION_COMPARTMENT_OCID".
            If "NB_SESSION_COMPARTMENT_OCID" is not set, the root compartment ID will be used.
            An ValueError will be raised if root compartment ID cannot be determined.
        project_id : str
            Target project to list deployments from.
            Defaults to the project id in the environment variable "PROJECT_OCID".
        kwargs :
            The values are passed to oci.data_science.DataScienceClient.list_model_deployments.

        Returns
        -------
        list
            A list of ModelDeployment objects.
        """
        deployments = OCIDataScienceModelDeployment.list(
            status=status,
            compartment_id=compartment_id,
            project_id=project_id,
            **kwargs,
        )
        return [cls()._update_from_oci_model(deployment) for deployment in deployments]

    @classmethod
    def list_df(
        cls,
        status: str = None,
        compartment_id: str = None,
        project_id: str = None,
    ) -> pd.DataFrame:
        """Returns the model deployments associated with current compartment and status
            as a Dataframe that can be easily visualized

        Parameters
        ----------
        status : str
            Status of deployment. Defaults to None.
            Allowed values: `ACTIVE`, `CREATING`, `DELETED`, `DELETING`, `FAILED`, `INACTIVE` and `UPDATING`.
        compartment_id : str
            Target compartment to list deployments from.
            Defaults to the compartment set in the environment variable "NB_SESSION_COMPARTMENT_OCID".
            If "NB_SESSION_COMPARTMENT_OCID" is not set, the root compartment ID will be used.
            An ValueError will be raised if root compartment ID cannot be determined.
        project_id : str
            Target project to list deployments from.
            Defaults to the project id in the environment variable "PROJECT_OCID".

        Returns
        -------
        DataFrame
            pandas Dataframe containing information about the ModelDeployments
        """
        model_deployments = cls.list(
            status=status, compartment_id=compartment_id, project_id=project_id
        )
        if isinstance(status, str) or status == None:
            status = State._from_str(status)
        display = pd.DataFrame()
        ids, urls, status_list = [], [], []
        for model_deployment in model_deployments:
            state_of_model = State._from_str(model_deployment.lifecycle_state)
            if status == State.UNKNOWN or status.name == state_of_model.name:
                ids.append(model_deployment.model_deployment_id)
                urls.append(model_deployment.url)
                status_list.append(model_deployment.lifecycle_state)
        display["deployment_id"] = ids
        display["deployment_url"] = urls
        display["current_state"] = status_list
        return display

    @classmethod
    def from_id(cls, id: str) -> "ModelDeployment":
        """Loads the model deployment instance from ocid.

        Parameters
        ----------
        id: str
            The ocid of model deployment.

        Returns
        -------
        ModelDeployment
            The ModelDeployment instance (self).
        """
        return cls()._update_from_oci_model(OCIDataScienceModelDeployment.from_id(id))

    @classmethod
    def from_dict(cls, obj_dict: Dict) -> "ModelDeployment":
        """Loads model deployment instance from a dictionary of configurations.

        Parameters
        ----------
        obj_dict: Dict
            A dictionary of configurations.

        Returns
        -------
        ModelDeployment
            The model deployment instance.
        """
        if not isinstance(obj_dict, dict):
            raise ValueError(
                "The config data for initializing the model deployment is invalid."
            )
        spec = ads_utils.batch_convert_case(
            copy.deepcopy(obj_dict.get("spec")), "camel"
        )

        mappings = {
            cls.CONST_INFRASTRUCTURE: {
                MODEL_DEPLOYMENT_INFRASTRUCTURE_TYPE: ModelDeploymentInfrastructure,
            },
            cls.CONST_RUNTIME: {
                ModelDeploymentRuntimeType.CONDA: ModelDeploymentCondaRuntime,
                ModelDeploymentRuntimeType.CONTAINER: ModelDeploymentContainerRuntime,
            },
        }
        model_deployment = cls()

        for key, value in spec.items():
            if key in mappings:
                mapping = mappings[key]
                child_config = value
                if child_config.get("type") not in mapping:
                    raise NotImplementedError(
                        f"{key.title()} type: {child_config.get('type')} is not supported."
                    )
                model_deployment.set_spec(
                    key, mapping[child_config.get("type")].from_dict(child_config)
                )
            else:
                model_deployment.set_spec(key, value)

        return model_deployment

    def to_dict(self) -> Dict:
        """Serializes model deployment to a dictionary.

        Returns
        -------
        dict
            The model deployment serialized as a dictionary.
        """
        spec = copy.deepcopy(self._spec)
        for key, value in spec.items():
            if hasattr(value, "to_dict"):
                value = value.to_dict()
            spec[key] = value

        return {
            "kind": self.kind,
            "type": self.type,
            "spec": ads_utils.batch_convert_case(spec, "camel"),
        }

    def _update_from_oci_model(self, oci_model_instance) -> "ModelDeployment":
        """Updates model deployment instance from OCIDataScienceModelDeployment.

        Parameters
        ----------
        oci_model_instance: OCIDataScienceModelDeployment
            The OCIDataScienceModelDeployment instance.

        Returns
        -------
        ModelDeployment
            The model deployment instance.
        """
        self.dsc_model_deployment = oci_model_instance
        for key, value in self.attribute_map.items():
            if hasattr(oci_model_instance, value):
                self.set_spec(key, getattr(oci_model_instance, value))

        infrastructure = ModelDeploymentInfrastructure()
        self._extract_from_oci_model(
            infrastructure, oci_model_instance, infrastructure.sub_level_attribute_maps
        )

        model_deployment_configuration_details = getattr(
            oci_model_instance, "model_deployment_configuration_details", None
        )
        environment_configuration_details = getattr(
            model_deployment_configuration_details,
            "environment_configuration_details",
            None,
        )
        runtime = (
            ModelDeploymentContainerRuntime()
            if getattr(
                environment_configuration_details,
                "environment_configuration_type",
                None,
            )
            == OCIModelDeploymentRuntimeType.CONTAINER
            else ModelDeploymentCondaRuntime()
        )

        self._extract_from_oci_model(runtime, oci_model_instance)
        infrastructure.set_spec(
            infrastructure.CONST_WEB_CONCURRENCY,
            runtime.env.get("WEB_CONCURRENCY", None),
        )

        self.set_spec(self.CONST_INFRASTRUCTURE, infrastructure)
        self.set_spec(self.CONST_RUNTIME, runtime)

        return self

    @staticmethod
    def _extract_from_oci_model(
        dsc_instance: Union[ModelDeploymentInfrastructure, ModelDeploymentRuntime],
        oci_model_instance: OCIDataScienceModelDeployment,
        sub_level: Dict = {},
    ) -> Union[ModelDeploymentInfrastructure, ModelDeploymentRuntime]:
        """Extract attributes from OCIDataScienceModelDeployment.

        Parameters
        ----------
        dsc_instance: Union[ModelDeploymentInfrastructure, ModelDeploymentRuntime]
            The ModelDeploymentInfrastructure or ModelDeploymentRuntime instance.
        oci_model_instance: OCIDataScienceModelDeployment
            The OCIDataScienceModelDeployment instance.
        sub_level: Dict
            The sub level attribute maps of ModelDeploymentInfrastructure or ModelDeploymentRuntime

        Returns
        -------
        Union[ModelDeploymentInfrastructure, ModelDeploymentRuntime]
            The ModelDeploymentInfrastructure or ModelDeploymentRuntime instance.
        """
        for infra_attr, dsc_attr in dsc_instance.payload_attribute_map.items():
            value = get_value(oci_model_instance, dsc_attr)
            if value:
                if infra_attr not in sub_level:
                    dsc_instance._spec[infra_attr] = value
                else:
                    dsc_instance._spec[infra_attr] = {}
                    for sub_infra_attr, sub_dsc_attr in sub_level[infra_attr].items():
                        sub_value = get_value(value, sub_dsc_attr)
                        if sub_value:
                            dsc_instance._spec[infra_attr][sub_infra_attr] = sub_value
        return dsc_instance

    def _build_model_deployment_details(self) -> CreateModelDeploymentDetails:
        """Builds CreateModelDeploymentDetails from model deployment instance.

        Returns
        -------
        CreateModelDeploymentDetails
            The CreateModelDeploymentDetails instance.
        """
        if not (self.infrastructure and self.runtime):
            raise ValueError(
                "Missing parameter runtime or infrastructure. Try reruning it after parameters are fully configured."
            )

        create_model_deployment_details = {
            self.CONST_DISPLAY_NAME: self.display_name or self._random_display_name(),
            self.CONST_DESCRIPTION: self.description,
            self.CONST_DEFINED_TAG: self.defined_tags,
            self.CONST_FREEFORM_TAG: self.freeform_tags,
            self.runtime.CONST_DEPLOYMENT_MODE: self.runtime.deployment_mode,
            self.infrastructure.CONST_COMPARTMENT_ID: self.infrastructure.compartment_id
            or COMPARTMENT_OCID,
            self.infrastructure.CONST_PROJECT_ID: self.infrastructure.project_id
            or PROJECT_OCID,
            self.infrastructure.CONST_MODEL_DEPLOYMENT_CONFIG_DETAILS: self._build_model_deployment_configuration_details(),
            self.infrastructure.CONST_CATEGORY_LOG_DETAILS: self._build_category_log_details(),
        }

        return OCIDataScienceModelDeployment(
            **create_model_deployment_details
        ).to_oci_model(CreateModelDeploymentDetails)

    def _update_model_deployment_details(self) -> UpdateModelDeploymentDetails:
        """Builds UpdateModelDeploymentDetails from model deployment instance.

        Returns
        -------
        UpdateModelDeploymentDetails
            The UpdateModelDeploymentDetails instance.
        """
        if not (self.infrastructure and self.runtime):
            raise ValueError(
                "Missing parameter runtime or infrastructure. Try reruning it after parameters are fully configured."
            )

        update_model_deployment_details = {
            self.CONST_DISPLAY_NAME: self.display_name,
            self.CONST_DESCRIPTION: self.description,
            self.CONST_DEFINED_TAG: self.defined_tags,
            self.CONST_FREEFORM_TAG: self.freeform_tags,
            self.infrastructure.CONST_MODEL_DEPLOYMENT_CONFIG_DETAILS: self._build_model_deployment_configuration_details(),
            self.infrastructure.CONST_CATEGORY_LOG_DETAILS: self._build_category_log_details(),
        }
        return OCIDataScienceModelDeployment(
            **update_model_deployment_details
        ).to_oci_model(UpdateModelDeploymentDetails)

    def _build_model_deployment_configuration_details(self) -> Dict:
        """Builds model deployment configuration details from model deployment instance.

        Returns
        -------
        Dict:
            Dict contains model deployment configuration details.
        """
        infrastructure = self.infrastructure
        runtime = self.runtime

        instance_configuration = {
            infrastructure.CONST_INSTANCE_SHAPE_NAME: infrastructure.shape_name,
            infrastructure.CONST_MODEL_DEPLOYMENT_INSTANCE_SHAPE_CONFIG_DETAILS: infrastructure.shape_config_details,
        }

        scaling_policy = {
            infrastructure.CONST_POLICY_TYPE: "FIXED_SIZE",
            infrastructure.CONST_INSTANCE_COUNT: infrastructure.replica,
        }

        if not runtime.model_uri:
            raise ValueError(
                "Missing parameter model uri. Try reruning it after model uri is configured."
            )

        model_id = runtime.model_uri
        if not model_id.startswith("ocid"):
            model_id = OCIClientManager().prepare_artifact(
                model_uri=runtime.model_uri,
                properties=dict(
                    display_name=self.display_name,
                    compartment_id=self.infrastructure.compartment_id
                    or COMPARTMENT_OCID,
                    project_id=self.infrastructure.project_id or PROJECT_OCID,
                ),
            )

        model_configuration_details = {
            infrastructure.CONST_BANDWIDTH_MBPS: infrastructure.bandwidth_mbps,
            infrastructure.CONST_INSTANCE_CONFIG: instance_configuration,
            runtime.CONST_MODEL_ID: model_id,
            infrastructure.CONST_SCALING_POLICY: scaling_policy,
        }

        environment_variables = runtime.env
        if infrastructure.web_concurrency:
            environment_variables["WEB_CONCURRENCY"] = str(
                infrastructure.web_concurrency
            )
            runtime.set_spec(runtime.CONST_ENV, environment_variables)
        environment_configuration_details = {
            runtime.CONST_ENVIRONMENT_CONFIG_TYPE: runtime.environment_config_type,
            runtime.CONST_ENVIRONMENT_VARIABLES: runtime.env,
        }

        if runtime.environment_config_type == OCIModelDeploymentRuntimeType.CONTAINER:
            if not hasattr(
                oci.data_science.models,
                "OcirModelDeploymentEnvironmentConfigurationDetails",
            ):
                raise EnvironmentError(
                    "Container runtime hasn't been supported in the current OCI SDK installed."
                )
            environment_configuration_details["image"] = runtime.image
            environment_configuration_details["imageDigest"] = runtime.image_digest
            environment_configuration_details["cmd"] = runtime.cmd
            environment_configuration_details["entrypoint"] = runtime.entrypoint
            environment_configuration_details["serverPort"] = runtime.server_port
            environment_configuration_details[
                "healthCheckPort"
            ] = runtime.health_check_port

        model_deployment_configuration_details = {
            infrastructure.CONST_DEPLOYMENT_TYPE: "SINGLE_MODEL",
            infrastructure.CONST_MODEL_CONFIG_DETAILS: model_configuration_details,
            runtime.CONST_ENVIRONMENT_CONFIG_DETAILS: environment_configuration_details,
        }

        if runtime.deployment_mode == ModelDeploymentMode.STREAM:
            if not hasattr(oci.data_science.models, "StreamConfigurationDetails"):
                raise EnvironmentError(
                    "Model deployment mode hasn't been supported in the current OCI SDK installed."
                )
            model_deployment_configuration_details[
                infrastructure.CONST_STREAM_CONFIG_DETAILS
            ] = {
                runtime.CONST_INPUT_STREAM_IDS: runtime.input_stream_ids,
                runtime.CONST_OUTPUT_STREAM_IDS: runtime.output_stream_ids,
            }

        return model_deployment_configuration_details

    def _build_category_log_details(self) -> Dict:
        """Builds category log details from model deployment instance.

        Returns
        -------
        Dict:
            Dict contains category log details.
        """
        if self.infrastructure.log_group_id and self.infrastructure.log_id:
            log_group_details = {
                self.infrastructure.CONST_LOG_GROUP_ID: self.infrastructure.log_group_id,
                self.infrastructure.CONST_LOG_ID: self.infrastructure.log_id,
            }
            return {
                self.infrastructure.CONST_ACCESS: log_group_details,
                self.infrastructure.CONST_PREDICT: log_group_details,
            }

        return {
            self.infrastructure.CONST_ACCESS: {
                self.infrastructure.CONST_LOG_GROUP_ID: self.infrastructure.access_log.get(
                    "logGroupId", None
                ),
                self.infrastructure.CONST_LOG_ID: self.infrastructure.access_log.get(
                    "logId", None
                ),
            },
            self.infrastructure.CONST_PREDICT: {
                self.infrastructure.CONST_LOG_GROUP_ID: self.infrastructure.predict_log.get(
                    "logGroupId", None
                ),
                self.infrastructure.CONST_LOG_ID: self.infrastructure.predict_log.get(
                    "logId", None
                ),
            },
        }

    def _random_display_name(self):
        """Generates a random display name."""
        return f"{self._PREFIX}-{ads_utils.get_random_name_for_resource()}"

    def _extract_spec_kwargs(self, **kwargs) -> Dict:
        """Extract spec related keyword arguments from kwargs.

        Parameters
        ----------
        kwargs

        Returns
        -------
        Dict:
            Dict contains model deployment spec related keyword arguments.
        """
        spec_kwargs = {}
        for attribute in self.initialize_spec_attributes:
            if attribute in kwargs:
                spec_kwargs[attribute] = kwargs[attribute]
        return spec_kwargs



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from distutils import dir_util
import os
import shutil
from collections.abc import Iterable

import numpy as np
import pandas as pd
from ads.common import logger, utils
from ads.common.model_export_util import (
    Progress_Steps_W_Fn,
    Progress_Steps_Wo_Fn,
    prepare_generic_model,
    serialize_model,
)
from ads.model.transformer.onnx_transformer import ONNXTransformer
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.common.decorator.deprecate import deprecated
from ads.common.utils import is_notebook
from ads.dataset.pipeline import TransformerPipeline
from sklearn.pipeline import Pipeline

Unsupported_Model_Types = []
NoTransformModels = ["torch", "tensorflow", "keras", "automl"]


class ADSModel(object):
    def __init__(
        self,
        est,
        target=None,
        transformer_pipeline=None,
        client=None,
        booster=None,
        classes=None,
        name=None,
    ):
        """
        Construct an ADSModel

        Parameters
        ----------
        est: fitted estimator object
            The estimator can be a standard sklearn estimator, a keras, lightgbm, or xgboost estimator, or any other object that implement methods from
            (BaseEstimator, RegressorMixin) for regression or (BaseEstimator, ClassifierMixin) for classification.
        target: PandasSeries
            The target column you are using in your dataset, this is assigned as the "y" attribute.
        transformer_pipeline: TransformerPipeline
            A custom trasnformer pipeline object.
        client: Str
            Currently unused.
        booster: Str
            Currently unused.
        classes: list, optional
            List of target classes. Required for classification problem if the est does not contain classes_ attribute.
        name: str, optional
            Name of the model.
        """
        self.est = est
        if utils.is_same_class(transformer_pipeline, Pipeline):
            self.transformer_pipeline = TransformerPipeline(transformer_pipeline.steps)
        elif isinstance(transformer_pipeline, list):
            self.transformer_pipeline = TransformerPipeline(transformer_pipeline)
        else:
            self.transformer_pipeline = transformer_pipeline
        self.target = target
        if classes is not None:
            self.classes_ = classes
        self.name = (
            name if name is not None else str(est)
        )  # Let the estimator define its own representation
        # These parameters make sense for dask_xgboost
        self.client = client
        self.booster = booster
        self._get_underlying_model_type()

    @staticmethod
    def from_estimator(est, transformers=None, classes=None, name=None):
        """
        Build ADSModel from a fitted estimator

        Parameters
        ----------
        est: fitted estimator object
            The estimator can be a standard sklearn estimator or any object that implement methods from
            (BaseEstimator, RegressorMixin) for regression or (BaseEstimator, ClassifierMixin) for classification.
        transformers: a scalar or an iterable of objects implementing transform function, optional
            The transform function would be applied on data before calling predict and predict_proba on estimator.
        classes: list, optional
            List of target classes. Required for classification problem if the est does not contain classes_ attribute.
        name: str, optional
            Name of the model.

        Returns
        -------
        model: ads.common.model.ADSModel
        Examples
        --------
        >>> model = MyModelClass.train()
        >>> model_ads = from_estimator(model)
        """
        if hasattr(est, "predict"):
            return ADSModel(
                est, transformer_pipeline=transformers, classes=classes, name=name
            )
        elif callable(est):
            return ADSModel(
                est, transformer_pipeline=transformers, classes=classes, name=name
            )

    # determine if the model is one of the common types without importing all of the libraries
    def _get_underlying_model_type(self):
        # WARNING!! Do not change the order. Particularly, sklearn must be last, as many classes extend sklearn
        model_bases = utils.get_base_modules(self.est)
        # should we be going past the first pkg name??
        if any([str(x)[:15] == "<class 'automl." for x in model_bases]):
            self._underlying_model = (
                "automl"  # always has "automl.interface.pipeline.Pipeline" ?
            )
        elif any([str(x)[:12] == "<class 'h2o." for x in model_bases]):
            self._underlying_model = (
                "h2o"  # always has "h2o.model.model_base.ModelBase" ?
            )
        elif any([str(x)[:17] == "<class 'lightgbm." for x in model_bases]):
            self._underlying_model = (
                "lightgbm"  # either "lightgbm.sklearn.LGBMModel" or "lightgbm.Booster"
            )
        elif any([str(x)[:16] == "<class 'xgboost." for x in model_bases]):
            self._underlying_model = (
                "xgboost"  # always has "xgboost.sklearn.XGBModel" or "xgboost."
            )
        elif any([str(x)[:14] == "<class 'torch." for x in model_bases]):
            self._underlying_model = "torch"  # "torch.nn.modules.module.Module"
        elif any([str(x)[:14] == "<class 'mxnet." for x in model_bases]):
            self._underlying_model = "mxnet"
        elif any([str(x)[:19] == "<class 'tensorflow." for x in model_bases]):
            self._underlying_model = (
                "tensorflow"  # "tensorflow.python.module.module.Module"
            )
            if any(
                [str(x)[:32] == "<class 'tensorflow.python.keras." for x in model_bases]
            ):
                self._underlying_model = "keras"
        elif any([str(x)[:13] == "<class 'pyod." for x in model_bases]):
            self._underlying_model = "pyod"  # always has pyod.models.base.BaseDetector
        elif any([str(x)[:16] == "<class 'sklearn." for x in model_bases]):
            self._underlying_model = (
                "sklearn"  # always has "sklearn.base.BaseEstimator"
            )
        else:
            self._underlying_model = "Unknown"
        return

    def rename(self, name):
        """
        Changes the name of a model

        Parameters
        ----------
        name: str
            A string which is supplied for naming a model.
        """
        self.name = name

    @deprecated(
        "2.6.6",
        details="Use framework specific Model utility class for saving and deploying model. Check https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_registration/quick_start.html",
    )
    def predict(self, X):
        """
        Runs the models predict function on some data

        Parameters
        ----------
        X: ADSData
            A ADSData object which holds the examples to be predicted on.

        Returns
        -------
        Union[List, pandas.Series], depending on the estimator
            Usually a list or PandasSeries of predictions
        """
        X = self.transform(X)
        if self._underlying_model in ["torch"]:
            return self.est(X)
        if self.client is not None and self.booster is not None:
            return self.est.predict(self.client, self.booster, X).persist()
        else:
            return self.est.predict(X)

    # For callable estimators, this will be more natural for ADSModel to support
    __call__ = predict

    def predict_proba(self, X):
        """
        Runs the models predict probabilities function on some data

        Parameters
        ----------
        X: ADSData
            A ADSData object which holds the examples to be predicted on.

        Returns
        -------
        Union[List, pandas.Series], depending on the estimator
            Usually a list or PandasSeries of predictions
        """
        X = self.transform(X)
        if self._underlying_model in ["torch"]:
            return self.est(X)
        if self.client is not None and self.booster is not None:
            return self.est.predict_proba(self.client, self.booster, X).persist()
        else:
            return self.est.predict_proba(X)

    @deprecated(
        "2.6.6",
        details="Use framework specific Model utility class for saving and deploying model. Check https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_registration/quick_start.html",
    )
    def score(self, X, y_true, score_fn=None):
        """
        Scores a model according to a custom score function

        Parameters
        ----------
        X: ADSData
            A ADSData object which holds the examples to be predicted on.
        y_true: ADSData
            A ADSData object which holds ground truth labels for the examples which are being predicted on.
        score_fn: Scorer (callable)
            A callable object that returns a score, usually created with sklearn.metrics.make_scorer().

        Returns
        -------
        float, depending on the estimator
            Almost always a scalar score (usually a float).
        """
        X = self.transform(X)
        if score_fn:
            return score_fn(self, X, y_true)
        else:
            assert hasattr(self.est, "score"), (
                f"Could not find a score function for estimator of type: "
                f"{self._underlying_model}. Pass in your desired scoring "
                f"function to score_fn "
            )
            if self.client is not None and self.booster is not None:
                return self.est.score(self.client, self.booster, X, y_true).persist()
            else:
                return self.est.score(X, y_true)

    @deprecated(
        "2.6.6",
        details="Use framework specific Model utility class for saving and deploying model. Check https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_registration/quick_start.html",
    )
    def summary(self):
        """
        A summary of the ADSModel
        """
        print(self)

    def __repr__(self):
        if self._underlying_model == "automl":
            framework = self.est.pipeline.trained_model.__class__.__module__
            est = self.est.selected_model_
            params = self.est.selected_model_params_
        else:
            framework = self.est.__class__.__module__
            est = self.est.__class__.__name__
            params = self.est.get_params() if hasattr(self.est, "get_params") else None
        return (
            "Framework: %s\n" % framework
            + "Estimator class: %s\n" % est
            + "Model Parameters: %s\n" % params
        )

    def __getattr__(self, item):
        return getattr(self.est, item)

    @deprecated(
        "2.6.6",
        details="Use framework specific Model utility class for saving and deploying model. Check https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_registration/quick_start.html",
    )
    def transform(self, X):
        """
        Process some ADSData through the selected ADSModel transformers

        Parameters
        ----------
        X: ADSData
            A ADSData object which holds the examples to be transformed.
        """

        if hasattr(X, "copy"):
            X = X.copy()
        if self.transformer_pipeline is not None:
            transformer_pipeline = self.transformer_pipeline
            if not isinstance(transformer_pipeline, Iterable):
                transformer_pipeline = [self.transformer_pipeline]
            for transformer in transformer_pipeline:
                try:
                    X = transformer.transform(X)
                except Exception as e:
                    pass
                    # logger.warn("Skipping pre-processing.")
        if self.target is not None and self.target in X.columns:
            X = X.drop(self.target, axis=1)
        return X

    def is_classifier(self):
        """
        Returns True if ADS believes that the model is a classifier

        Returns
        -------
        Boolean: True if the model is a classifier, False otherwise.
        """
        return hasattr(self, "classes_") and self.classes_ is not None

    @deprecated(
        "2.6.6",
        details="Use framework specific Model utility class for saving and deploying model. Check https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_registration/quick_start.html",
    )
    def feature_names(self, X=None):
        model_type = self._underlying_model
        if model_type == "sklearn":
            return X.columns
        elif model_type == "automl":
            return self.est.selected_features_names_
        elif model_type == "lightgbm":
            try:
                return self.est.feature_name()
            except AttributeError:
                return X.columns
        elif model_type == "torch":
            return []
        elif model_type == "xgboost":
            try:
                return self.est.feature_name()
            except AttributeError:
                return X.columns
        elif model_type == "tensorflow":
            return []
        elif model_type == "keras":
            return []
        elif model_type == "mxnet":
            return []
        else:
            try:
                return self.est.feature_names()
            except:
                logger.warning(
                    f"Could not find a model of type {model_type}. Therefore, "
                    f"there are no `feature_names`."
                )
                return []

    def _onnx_data_transformer(self, X, impute_values={}, **kwargs):
        if self._underlying_model in NoTransformModels:
            return X
        try:
            if hasattr(self, "onnx_data_preprocessor") and isinstance(
                self.onnx_data_preprocessor, ONNXTransformer
            ):
                return self.onnx_data_preprocessor.transform(X=X)

            self.onnx_data_preprocessor = ONNXTransformer()
            return self.onnx_data_preprocessor.fit_transform(
                X=X, impute_values=impute_values
            )
        except Exception as e:
            print(f"Warning: Onnx Data Transformation was unsuccessful with error: {e}")
            raise e

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__ = state

    @deprecated(
        "2.6.6",
        details="Use framework specific Model utility class for saving and deploying model. Check https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_registration/quick_start.html",
    )
    def prepare(
        self,
        target_dir=None,
        data_sample=None,
        X_sample=None,
        y_sample=None,
        include_data_sample=False,
        force_overwrite=False,
        fn_artifact_files_included=False,
        fn_name="model_api",
        inference_conda_env=None,
        data_science_env=False,
        ignore_deployment_error=False,
        use_case_type=None,
        inference_python_version=None,
        imputed_values={},
        **kwargs,
    ):
        """
        Prepare model artifact directory to be published to model catalog

        Parameters
        ----------
        target_dir : str, default: model.name[:12]
            Target directory under which the model artifact files need to be added
        data_sample : ADSData
            Note: This format is preferable to X_sample and y_sample.
            A sample of the test data that will be provided to predict() API of scoring script
            Used to generate schema_input.json and schema_output.json which defines the input and output formats
        X_sample : pandas.DataFrame
            A sample of input data that will be provided to predict() API of scoring script
            Used to generate schema.json which defines the input formats
        y_sample : pandas.Series
            A sample of output data that is expected to be returned by predict() API of scoring script,
            corresponding to X_sample
            Used to generate schema_output.json which defines the output formats
        force_overwrite : bool, default: False
            If True, overwrites the target directory if exists already
        fn_artifact_files_included : bool, default: True
            If True, generates artifacts to export a model as a function without ads dependency
        fn_name : str, default: 'model_api'
            Required parameter if fn_artifact_files_included parameter is setup.
        inference_conda_env : str, default: None
            Conda environment to use within the model deployment service for inferencing
        data_science_env : bool, default: False
            If set to True, datascience environment represented by the slug in the training conda environment will be used.
        ignore_deployment_error : bool, default: False
            If set to True, the prepare will ignore all the errors that may impact model deployment
        use_case_type: str
            The use case type of the model. Use it through UserCaseType class or string provided in UseCaseType. For
            example, use_case_type=UseCaseType.BINARY_CLASSIFICATION or use_case_type="binary_classification". Check
            with UseCaseType class to see all supported types.
        inference_python_version: str, default:None.
            If provided will be added to the generated runtime yaml

        **kwargs
        --------
        max_col_num: (int, optional). Defaults to utils.DATA_SCHEMA_MAX_COL_NUM.
            The maximum column size of the data that allows to auto generate schema.

        Returns
        -------
        model_artifact: an instance of `ModelArtifact` that can be used to test the generated scoring script
        """
        if include_data_sample:
            logger.warning(
                f"Parameter `include_data_sample` is deprecated and removed in future releases. "
                f"Data sample is not saved. You can manually save the data sample to {target_dir}."
            )
        # Add 2 for model and schema (Artifact Directory gets skipped in prepare_generic when progress is passed in).
        ProgressStepsWFn = Progress_Steps_W_Fn + 1
        ProgressStepsWoFn = Progress_Steps_Wo_Fn + 1
        if target_dir is None:
            logger.info(
                f"Using the default directory {self.name[:12]} "
                f"to create the model artifact. Use `target_dir` to specify a directory."
            )
        can_generate_fn_files = (
            fn_artifact_files_included
            and self._underlying_model not in Unsupported_Model_Types
        )
        assert data_sample is not None or X_sample is not None, (
            "You must provide a data sample to infer the input and output data types"
            " which are used when converting the the model to an equivalent onnx model. "
            "This can be done as an ADSData object with "
            "the parameter `data_sample`, or as X and y samples "
            "to X_sample and y_sample respectively. "
        )
        with utils.get_progress_bar(
            ProgressStepsWFn if can_generate_fn_files else ProgressStepsWoFn
        ) as progress:
            progress.update("Preparing Model Artifact Directory")
            if os.path.exists(target_dir):
                if not force_overwrite:
                    raise ValueError("Directory already exists, set force to overwrite")
            os.makedirs(target_dir, exist_ok=True)

            # Bring in model-ignore file
            shutil.copyfile(
                os.path.join(
                    os.path.dirname(os.path.realpath(__file__)),
                    "artifact/.model-ignore",
                ),
                os.path.join(target_dir, ".model-ignore"),
            )
            dir_util._path_created = {}

            progress.update("Serializing model")
            # Transform the data to be onnx-ready
            X_sample = (
                data_sample.X
                if X_sample is None and data_sample is not None
                else X_sample
            )
            y_sample = (
                data_sample.y
                if y_sample is None and data_sample is not None
                else y_sample
            )

            X_trans = self._onnx_data_transformer(
                X=X_sample, imputed_values=imputed_values
            )

            model_kwargs = serialize_model(
                model=self,
                target_dir=target_dir,
                X=X_trans,
                y=y_sample,
                model_type=self._underlying_model,
            )
            max_col_num = kwargs.get("max_col_num", utils.DATA_SCHEMA_MAX_COL_NUM)

            if self._underlying_model not in NoTransformModels:
                try:
                    self.onnx_data_preprocessor.save(
                        os.path.join(target_dir, "onnx_data_transformer.json")
                    )
                except Exception as e:
                    logger.error(
                        f"Unable to serialize the data transformer due to: {e}."
                    )
                    raise e

            if model_kwargs.get("serializer", "") != "onnx":
                model_kwargs["model_libs"] = utils.extract_lib_dependencies_from_model(
                    self.est
                )
            model_kwargs["underlying_model"] = self._underlying_model
            model_kwargs["progress"] = progress
            model_kwargs["inference_conda_env"] = inference_conda_env
            model_kwargs["data_science_env"] = data_science_env
            model_kwargs["ignore_deployment_error"] = ignore_deployment_error
            model_kwargs["use_case_type"] = use_case_type
            model_kwargs["max_col_num"] = max_col_num
            model_artifact = prepare_generic_model(
                target_dir,
                model=self.est,
                data_sample=data_sample,
                X_sample=X_sample,
                y_sample=y_sample,
                fn_artifact_files_included=fn_artifact_files_included,
                fn_name=fn_name,
                force_overwrite=force_overwrite,
                inference_python_version=inference_python_version,
                **model_kwargs,
            )
            try:
                model_file_name = (
                    "model.pkl" if self._underlying_model == "automl" else "model.onnx"
                )
                model_artifact.reload(model_file_name=model_file_name)
            except Exception as e:
                print(str(e))
                msg = (
                    "\nWARNING: Validation using scoring script failed. Update the inference script("
                    "score.py) as required. "
                )
                print("\033[93m" + msg + "\033[0m")

            # __pycache__ was created during model_artifact.reload() above
            if os.path.exists(os.path.join(target_dir, "__pycache__")):
                shutil.rmtree(
                    os.path.join(target_dir, "__pycache__"), ignore_errors=True
                )

            logger.info(model_artifact.__repr__())
            return model_artifact

    def visualize_transforms(self):
        """
        A graph of the ADSModel transformer pipeline.
        It is only supported in JupyterLabs Notebooks.
        """
        self.transformer_pipeline.visualize()

    @runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
    def show_in_notebook(self):
        """
        Describe the model by showing it's properties
        """
        # ['Is Regression', self.est.is_regression],
        if self._underlying_model == "automl":
            info = [
                ["Model Name", self.name],
                ["Target Variable", self.target],
                ["Selected Algorithm", self.est.selected_model_],
                ["Task", self.est.task],
                ["Training Dataset Size", self.est.train_shape_],
                ["CV", self.est.cv_],
                ["Optimization Metric", self.est.score_metric],
                ["Selected Hyperparameters", self.est.selected_model_params_],
                ["Initial Number of Features", self.est.train_shape_[1]],
                ["Initial Features", self.est.pipeline.orig_feature_names],
                ["Selected Number of Features", len(self.est.selected_features_names_)],
                ["Selected Features", self.est.selected_features_names_],
            ]
        else:
            info = [
                ["Model Name", self.name],
                [
                    "Target Variable",
                    self.target
                    if self.target is not None
                    else "not available from estimator",
                ],
                [
                    "Selected Hyperparameters",
                    self.est.get_params() if hasattr(self.est, "get_params") else None,
                ],
                ["Framework", self.est.__class__.__module__],
                ["Estimator Class", self.est.__class__.__name__],
                [
                    "Contained Estimator",
                    self.est.est.__class__.__name__
                    if hasattr(self.est, "est")
                    else None,
                ],
            ]
        info_df = pd.DataFrame(info)

        if is_notebook():
            with pd.option_context(
                "display.max_colwidth",
                1000,
                "display.width",
                None,
                "display.precision",
                4,
            ):
                from IPython.core.display import HTML, display

                display(HTML(info_df.to_html(index=False, header=False)))
        return info

    @staticmethod
    @runtime_dependency(module="skl2onnx", install_from=OptionalDependency.ONNX)
    def get_init_types(df, underlying_model=None):

        from skl2onnx.common.data_types import FloatTensorType

        if underlying_model == "sklearn":
            n_cols = len(df.columns)
            return [("input", FloatTensorType([None, n_cols]))], {"type": np.float32}
        return [], {}

    @staticmethod
    @runtime_dependency(module="skl2onnx", install_from=OptionalDependency.ONNX)
    def convert_dataframe_schema(df, drop=None):
        from skl2onnx.common.data_types import (
            FloatTensorType,
            Int64TensorType,
            StringTensorType,
        )

        inputs = []
        for k, v in zip(df.columns, df.dtypes):
            if drop is not None and k in drop:
                continue
            if v == "int64":
                t = Int64TensorType([1, 1])
            elif v == "float64":
                t = FloatTensorType([1, 1])
            else:
                t = StringTensorType([1, 1])
            inputs.append((k, t))
        return inputs



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from abc import abstractmethod
from typing import Dict


class ConfigProcessor:
    def __init__(self, config: Dict = None) -> None:
        """
        Initializing a ConfigProcessor object given a configuration.

        Parameters
        ----------
        config: dict
            a dictionary of configurations
        """
        self.config = config if config else {}

    @abstractmethod
    def process(self, **kwargs) -> "ConfigProcessor":
        """
        Perform some processing on configuration
        associated with this object.

        Parameters
        ----------
        kwargs: dict
            keyword arguments passed to the function

        Returns
        -------
        ConfigProcessor
            this instance itself
        """
        pass

    def step(self, cls, **kwargs) -> "ConfigProcessor":
        """
        Perform some processing according to the cls given.

        Parameters
        ----------
        cls: `ConfigProcessor`
            a subclass of `ConfigProcessor`
        kwargs: dict
            keyword arguments passed to the `process` function of `cls`

        Returns
        -------
        `ConfigProcessor`
            this instance itself
        """
        return cls(self.config).process(**kwargs)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


class Timeseries:
    def __init__(self, col_name, df, date_range=None, min=None, max=None):
        self.col_name = col_name
        self.df = df
        self.date_range = date_range
        self.min = min
        self.max = max

    def plot(self, **kwargs):
        # this could be improved :)
        pass



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from abc import ABC, abstractmethod
from typing import Any


class Parser(ABC):
    """Data Parser Interface."""

    @abstractmethod
    def parse(self) -> Any:
        pass



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from abc import abstractmethod
from typing import Dict

from ads.common.auth import get_signer


class Backend:
    """Interface for backend"""

    def __init__(self, config: Dict) -> None:
        self.config = config
        self.oci_auth = get_signer(
            config["execution"].get("oci_config", None),
            config["execution"].get("oci_profile", None),
        )
        self.profile = config["execution"].get("oci_profile", None)

    @abstractmethod
    def run(self) -> Dict:
        """
        Initiate a run.

        Returns
        -------
        None
        """

    def delete(self) -> None:
        """
        Delete a remote run.

        Returns
        -------
        None
        """

    def watch(self) -> None:
        """
        Stream logs from a remote run.

        Returns
        -------
        None
        """

    def cancel(self) -> None:
        """
        Cancel a remote run.

        Returns
        -------
        None
        """

    def apply(self) -> None:
        """
        Initiate Data Science service from YAML.

        Returns
        -------
        None
        """

    def activate(self) -> None:
        """
        Activate a remote service.

        Returns
        -------
        None
        """
        raise NotImplementedError("`activate` has not been implemented yet.")

    def deactivate(self) -> None:
        """
        Deactivate a remote service.

        Returns
        -------
        None
        """
        raise NotImplementedError("`deactivate` has not been implemented yet.")

    def run_diagnostics(self):
        """
        Implement Diagnostics check appropriate for the backend
        """



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and its affiliates.
# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from .model_deployer import ModelDeployer
from .model_deployment import (
    DEFAULT_POLL_INTERVAL,
    DEFAULT_WAIT_TIME,
    ModelDeployment,
)
from .model_deployment_properties import ModelDeploymentProperties



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import datetime
from typing import Optional

import oci.data_science
from ads.common.oci_datascience import OCIDataScienceMixin
from ads.common.oci_mixin import (
    LIFECYCLE_STOP_STATE,
    OCIModelNotExists,
    OCIModelWithNameMixin,
    OCIWorkRequestMixin,
)

from ads.config import COMPARTMENT_OCID

from oci.data_science.models import (
    CreateModelVersionSetDetails,
    UpdateModelVersionSetDetails,
)


class ModelVersionSetNotExists(Exception):
    pass


class ModelVersionSetNotSaved(Exception):
    pass


class DataScienceModelVersionSet(
    OCIDataScienceMixin,
    OCIModelWithNameMixin,
    OCIWorkRequestMixin,
    oci.data_science.models.ModelVersionSet,
):
    """Represents an OCI Data Science Model Version Set
    This class contains all attributes of the oci.data_science.models.ModelVersionSet
    The main purpose of this class is to link the `oci.data_science.models.ModelVersionSet` model
    and the related client methods. Mainly, linking the ModelVersionSet model (payload) to
    Create/Update/Get/List/Delete methods.

    A `DataScienceModelVersionSet` can be initialized by unpacking the properties stored in a dictionary (payload):

    .. code-block:: python

        properties = {
            "name": "experiment1",
            "description": "my experiment"
        }
        experiment = DataScienceModelVersionSet(**properties)

    The properties can also be OCI REST API payload, in which the keys are in camel format.

    .. code-block:: python

        payload = {
            "Id": "<model_version_set_ocid>",
            "compartmentId": "<compartment_ocid>",
            "Name": "<model_version_set_name>",
        }
        experiment = DataScienceModelVersionSet(**payload)
    """

    def create(self) -> "DataScienceModelVersionSet":
        """Creates model version set on OCI Data Science platform

        Returns
        -------
        DataScienceModelVersionSet
            The DataScienceModelVersionSet instance (self), which allows chaining additional method.
        """
        if not self.name:
            timestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M")
            self.name = f"model-version-set-{timestamp}"

        if not self.compartment_id:
            raise ValueError(
                "`compartment_id` must be specified for the model version set."
            )

        if not self.project_id:
            raise ValueError(
                "`project_id` must be specified for the model version set."
            )

        res = self.client.create_model_version_set(
            self.to_oci_model(CreateModelVersionSetDetails)
        )
        self.update_from_oci_model(res.data)

        return self

    def update(self) -> "DataScienceModelVersionSet":
        """Updates the Data Science Model Version Set."""
        res = self.client.update_model_version_set(
            self.id, self.to_oci_model(UpdateModelVersionSetDetails)
        )
        self.update_from_oci_model(res.data)

        return self

    def delete(
        self, delete_model: Optional[bool] = False
    ) -> "DataScienceModelVersionSet":
        """Deletes the model version set.

        Parameters
        ----------
        delete_model: (bool, optional). Defaults to False.
            By default, this parameter is false. A model version set can only be
            deleted if all the models associate with it are already in the DELETED state.
            You can optionally specify the deleteRelatedModels boolean query parameters to
            true, which deletes all associated models for you.

        Returns
        -------
        DataScienceModelVersionSet
            The DataScienceModelVersionSet instance (self), which allows chaining additional method.
        """
        response = self.client.delete_model_version_set(
            self.id, is_delete_related_models=delete_model
        )
        self.get_work_request_response(
            response,
            wait_for_state=LIFECYCLE_STOP_STATE,
            success_state="SUCCEEDED",
            wait_interval_seconds=1,
        )
        self.sync()
        return self

    @classmethod
    def from_ocid(cls, ocid: str) -> "DataScienceModelVersionSet":
        """Gets a Model Version Set by OCID.

        Parameters
        ----------
        ocid: str
            The OCID of the model version set.

        Returns
        -------
        DataScienceModelVersionSet
            An instance of DataScienceModelVersionSet.
        """
        return super().from_ocid(ocid)

    @classmethod
    def from_name(
        cls, name: str, compartment_id: Optional[str] = None
    ) -> "DataScienceModelVersionSet":
        """Gets a Model Version Set by name.

        Parameters
        ----------
        compartment_id: (str, optional). Defaults to None.
            Compartment OCID of the OCI resources. If `compartment_id` is not specified,
            the value will be taken from environment variables.
        name: str
            The name of the model version set.

        Returns
        -------
        DataScienceModelVersionSet
            An instance of DataScienceModelVersionSet.
        """
        compartment_id = compartment_id or COMPARTMENT_OCID
        try:
            return super().from_name(name=name, compartment_id=compartment_id)
        except OCIModelNotExists:
            raise ModelVersionSetNotExists(f"The model version set `{name}` not found.")



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2020, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from __future__ import absolute_import, division, print_function

import copy

import numpy as np
import pandas as pd
import scipy

from ads.common import logger, utils
from ads.common.decorator.deprecate import deprecated
from ads.common.model import ADSModel
from ads.dataset import helper
from ads.dataset.classification_dataset import (
    BinaryClassificationDataset,
    BinaryTextClassificationDataset,
    MultiClassClassificationDataset,
    MultiClassTextClassificationDataset,
)
from ads.dataset.dataset_with_target import ADSDatasetWithTarget
from ads.dataset.pipeline import TransformerPipeline
from ads.dataset.regression_dataset import RegressionDataset
from ads.type_discovery.type_discovery_driver import TypeDiscoveryDriver
from ads.type_discovery.typed_feature import (
    ContinuousTypedFeature,
    DiscreteTypedFeature,
)

dataset_task_map = {
    BinaryClassificationDataset: utils.ml_task_types.BINARY_CLASSIFICATION,
    BinaryTextClassificationDataset: utils.ml_task_types.BINARY_TEXT_CLASSIFICATION,
    MultiClassClassificationDataset: utils.ml_task_types.MULTI_CLASS_CLASSIFICATION,
    MultiClassTextClassificationDataset: utils.ml_task_types.MULTI_CLASS_TEXT_CLASSIFICATION,
    RegressionDataset: utils.ml_task_types.REGRESSION,
}


@deprecated(
    details="Working with AutoML has moved from within ADS to working directly with the AutoMLx library. AutoMLx are preinstalled in conda pack automlx_p38_cpu_v2 and later, and can now be updated independently of ADS. AutoMLx documentation may be found at https://docs.oracle.com/en-us/iaas/tools/automlx/latest/html/multiversion/v23.1.1/index.html. Notebook examples are in Oracle's samples repository: https://github.com/oracle-samples/oci-data-science-ai-samples/tree/master/notebook_examples and a migration tutorial can be found at https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_training/automl/quick_start.html .",
    raise_error=True,
)
def get_ml_task_type(X, y, classes):
    """
    Gets the ML task type and returns it.

    Parameters
    ----------
    X: Dataframe
        The training dataframe
    Y: Dataframe
        The testing dataframe
    Classes: List
        a list of classes

    Returns
    -------
    ml_task_type:
        A particular task type like `REGRESSION`, `MULTI_CLASS_CLASSIFICATION`...
    """
    target_type = TypeDiscoveryDriver().discover(y.name, y)
    if isinstance(target_type, DiscreteTypedFeature):
        if len(classes) == 2:
            if helper.is_text_data(X):
                ml_task_type = utils.ml_task_types.BINARY_TEXT_CLASSIFICATION
            else:
                ml_task_type = utils.ml_task_types.BINARY_CLASSIFICATION
        else:
            if helper.is_text_data(X):
                ml_task_type = utils.ml_task_types.MULTI_CLASS_TEXT_CLASSIFICATION
            else:
                ml_task_type = utils.ml_task_types.MULTI_CLASS_CLASSIFICATION
    elif isinstance(target_type, ContinuousTypedFeature):
        ml_task_type = utils.ml_task_types.REGRESSION
    else:
        raise TypeError(
            "AutoML for target type ({0}) is not yet available".format(
                target_type.meta_data["type"]
            )
        )
    return ml_task_type


class AutoML:
    @deprecated(
        details="Working with AutoML has moved from within ADS to working directly with the AutoMLx library. AutoMLx are preinstalled in conda pack automlx_p38_cpu_v2 and later, and can now be updated independently of ADS. AutoMLx documentation may be found at https://docs.oracle.com/en-us/iaas/tools/automlx/latest/html/multiversion/v23.1.1/index.html. Notebook examples are in Oracle's samples repository: https://github.com/oracle-samples/oci-data-science-ai-samples/tree/master/notebook_examples and a migration tutorial can be found at https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_training/automl/quick_start.html .",
        raise_error=True,
    )
    def __init__(
        self,
        training_data,
        validation_data=None,
        provider=None,
        baseline="dummy",
        client=None,
    ):
        """
        Creates an Automatic machine learning object.

        Parameters
        ----------
        training_data : `ADSData` instance
        validation_data : `ADSData` instance
        provider : None or object of ads.automl.provider.AutoMLProvider
            If None, the default OracleAutoMLProvider will be used to generate the model
        baseline: None, "dummy", or object of ads.common.model.ADSModel (Default is "dummy")

            - If None, than no baseline is created,
            - If "dummy", than the DummyClassifier or DummyRegressor are used
            - If Object, than whatever estimator is provided will be used.

            This estimator must include a part of its pipeline which does preprocessing
            to handle categorical data
        client:
            Dask Client to use (optional)

        Examples
        --------
        >>> train, test = ds.train_test_split()
        >>> olabs_automl = OracleAutoMLProvider()
        >>> model, baseline = AutoML(train, provider=olabs_automl).train()
        """
        from ads.automl.provider import BaselineAutoMLProvider, OracleAutoMLProvider

        if hasattr(training_data, "transformer_pipeline"):
            self.transformer_pipeline = training_data.transformer_pipeline
        else:
            self.transformer_pipeline = None

        if isinstance(training_data, ADSDatasetWithTarget):
            training_data, _ = training_data.train_test_split(test_size=0.0)

        if isinstance(validation_data, ADSDatasetWithTarget):
            validation_data, _ = validation_data.train_test_split(test_size=0.0)

        X = (
            training_data.X.compute()
            if utils._is_dask_dataframe(training_data.X)
            else training_data.X
        )
        y = (
            training_data.y.compute()
            if utils._is_dask_dataframe(training_data.y)
            or utils._is_dask_series(training_data.y)
            else training_data.y
        )
        self.X_shape = X.shape

        X_valid = None
        y_valid = None
        if validation_data is not None:
            X_valid = (
                training_data.X.compute()
                if utils._is_dask_dataframe(training_data.X)
                else training_data.X
            )
            y_valid = (
                training_data.y.compute()
                if utils._is_dask_dataframe(training_data.y)
                or utils._is_dask_series(training_data.y)
                else training_data.y
            )

        if isinstance(y, pd.DataFrame):
            if len(y.columns) != 1:
                raise ValueError("Data must be 1-dimensional.")
            y = y.squeeze()
        elif isinstance(y, np.ndarray):
            y = pd.Series(y)
        if y.name:
            self.target_name = str(y.name)
        else:
            y.name = str(0)
            self.target_name = str(0)

        if isinstance(X, np.ndarray):
            X = pd.DataFrame(X)
        elif isinstance(X, scipy.sparse.csr.csr_matrix):
            X = pd.DataFrame(X.todense())
        self.feature_names = X.columns.values
        self.client = client
        class_names = y.unique()

        if training_data.dataset_type in dataset_task_map:
            self.ml_task_type = dataset_task_map[training_data.dataset_type]
        else:
            self.ml_task_type = get_ml_task_type(X, y, class_names)

        # We have decided to use OracleAutoMLProvider as the default if provider
        # is not specified.
        if provider is None:
            provider = OracleAutoMLProvider()
        self.no_baseline = False

        self.automl_provider = provider
        if baseline == None:
            self.no_baseline = True
            self.baseline_provider = None
            self.automl_provider.setup(
                X,
                y,
                self.ml_task_type,
                X_valid=X_valid,
                y_valid=y_valid,
                class_names=class_names,
                client=self.client,
            )
        elif baseline == "dummy":
            self.baseline_provider = BaselineAutoMLProvider(None)
            for provider in [self.baseline_provider, self.automl_provider]:
                provider.setup(
                    X,
                    y,
                    self.ml_task_type,
                    X_valid=X_valid,
                    y_valid=y_valid,
                    class_names=class_names,
                    client=self.client,
                )
        else:
            self.baseline_provider = BaselineAutoMLProvider(baseline)
            for provider in [self.baseline_provider, self.automl_provider]:
                provider.setup(
                    X,
                    y,
                    self.ml_task_type,
                    X_valid=X_valid,
                    y_valid=y_valid,
                    class_names=class_names,
                    client=self.client,
                )

    def __getattr__(self, attr):
        # If the attribute is not in the AutoML object,
        # then it'll search to see if it exists in the provider
        err_msg = ""
        if attr.startswith("__") and attr.endswith("__"):
            raise AttributeError(
                "{} object has no attribute '{}'.".format(self.__class__.__name__, attr)
            )

        if (
            "automl_provider" in self.__dict__.keys()
            and self.automl_provider is not None
            and hasattr(self.automl_provider, attr)
        ):
            return getattr(self.automl_provider, attr)
        else:
            raise AttributeError(
                "{} object has no attribute '{}'; {}".format(
                    self.__class__.__name__, attr, err_msg
                )
            )

    def train(self, **kwargs):
        r"""
        Returns a fitted automl model and a fitted baseline model.

        Parameters
        ----------
        kwargs : dict, optional
            kwargs passed to provider's train method

        Returns
        -------
        model: object of ads.common.model.ADSModel
            the trained automl model
        baseline: object of ads.common.model.ADSModel
            the baseline model to compare
        Examples
        --------
        >>> train, test = ds.train_test_split()
        >>> olabs_automl = OracleAutoMLProvider()
        >>> model, baseline = AutoML(train, provider=olabs_automl).train()
        """

        time_budgeted = "time_budget" in kwargs
        avail_n_cores = utils.get_cpu_count()

        warn_params = [
            (10**5, 4, "VM.Standard.E2.4"),
            (10**6, 16, "VM.Standard.2.16"),
        ]

        # train using automl and baseline
        if self.no_baseline:
            self.automl_provider.train(**kwargs)
            if self.transformer_pipeline is not None:
                transformer_pipeline1 = copy.deepcopy(self.transformer_pipeline)
                for transformer in self.automl_provider.get_transformer_pipeline():
                    transformer_pipeline1.add(transformer)
            else:
                pipeline1 = self.automl_provider.get_transformer_pipeline()
                transformer_pipeline1 = (
                    TransformerPipeline(pipeline1) if pipeline1 is not None else None
                )
            model = ADSModel(
                self.automl_provider.est, self.target_name, transformer_pipeline1
            )
            return model
        else:
            for provider in [self.automl_provider, self.baseline_provider]:
                provider.train(**kwargs)
            if self.transformer_pipeline is not None:
                transformer_pipeline1 = copy.deepcopy(self.transformer_pipeline)
                transformer_pipeline2 = copy.deepcopy(self.transformer_pipeline)
                for transformer in self.automl_provider.get_transformer_pipeline():
                    transformer_pipeline1.add(transformer)
                for transformer in self.baseline_provider.get_transformer_pipeline():
                    transformer_pipeline2.add(transformer)
            else:
                pipeline1 = self.automl_provider.get_transformer_pipeline()
                pipeline2 = self.baseline_provider.get_transformer_pipeline()
                transformer_pipeline1 = (
                    TransformerPipeline(pipeline1) if pipeline1 is not None else None
                )
                transformer_pipeline2 = (
                    TransformerPipeline(pipeline2) if pipeline2 is not None else None
                )
            model = ADSModel(
                self.automl_provider.est, self.target_name, transformer_pipeline1
            )
            baseline = ADSModel(
                self.baseline_provider.est, self.target_name, transformer_pipeline2
            )
            return model, baseline



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from ads.common import utils as common_utils
from ads.jobs.builders.base import Builder
from ads.jobs.builders.runtimes.base import Runtime
import logging

logger = logging.getLogger(__name__)


class Infrastructure(Builder):
    """Base class for job infrastructure"""

    @property
    def kind(self) -> str:
        """Kind of the object to be stored in YAML. All runtimes will have "infrastructure" as kind.
        Subclass will have different types.
        """
        return "infrastructure"

    @property
    def name(self) -> str:
        """The name of the job from the infrastructure."""
        raise NotImplementedError()

    def _assert_runtime_compatible(self, runtime: Runtime) -> None:
        """
        Check if a runtime is compatible with an engine. Raise an error if not.

        Parameters
        ----------
        runtime: `Runtime`
            a runtime object

        Returns
        -------
        None
        """

    def create(self, runtime: Runtime, **kwargs):
        """
        Create/deploy a job on the infrastructure.

        Parameters
        ----------
        runtime: `Runtime`
            a runtime object
        kwargs: dict
            additional arguments

        """
        raise NotImplementedError()

    def run(
        self,
        name: str = None,
        args: str = None,
        env_var: dict = None,
        freeform_tags: dict = None,
        wait: bool = False,
    ):
        """Runs a job on the infrastructure.

        Parameters
        ----------
        name : str, optional
            The name of the job run, by default None
        args : str, optional
            Command line arguments for the job run, by default None.
        env_var : dict, optional
            Environment variable for the job run, by default None
        freeform_tags : dict, optional
            Freeform tags for the job run, by default None
        wait : bool, optional
            Indicate if this method should wait for the run to finish before it returns, by default False.
        """
        raise NotImplementedError()

    def delete(self):
        """Deletes a job from the infrastructure."""
        raise NotImplementedError()

    def update(self, runtime: Runtime):
        """Updates a job.

        Parameters
        ----------
        runtime
            a runtime object
        """
        raise NotImplementedError()

    def list_jobs(self, **kwargs) -> list:
        """
        List jobs from the infrastructure.

        Parameters
        ----------
        kwargs: keyword arguments for filtering the results

        Returns
        -------
        list
            list of infrastructure objects, each representing a job from the infrastructure.
        """
        raise NotImplementedError()


class RunInstance:
    _DETAILS_LINK = ""

    def create(self):
        """Create a RunInstance Object."""
        raise NotImplementedError()

    @property
    def status(self):
        """Return status of the run."""
        raise NotImplementedError()

    def watch(self):
        """Show logs of a run."""
        raise NotImplementedError()

    def delete(self):
        """Delete a run."""
        raise NotImplementedError()

    def cancel(self):
        """Cancel a run."""
        raise NotImplementedError()

    @property
    def run_details_link(self) -> str:
        """
        Link to run details page in OCI console

        Returns
        -------
        str
            The link to the details page in OCI console.
        """
        if not self._DETAILS_LINK:
            return ""

        try:
            return self._DETAILS_LINK.format(
                region=common_utils.extract_region(), id=self.id
            )
        except Exception as ex:
            print(str(ex))
            logger.info(
                "Error occurred in attempt to extract the link to the resource. "
                f"Details: {ex}"
            )
        return ""
